from typing import Callable, Union
from opendp.mod import Domain, Metric, _PartialConstructor, Measurement, Transformation
import inspect


def supports_partial(constructor: Callable[..., Union[Transformation, Measurement]]) -> bool:
    """Check if the constructor can be written as a then_ (partial application) function.

    This is true if the first two arguments are input_domain and input_metric.
    """
    arg_names = inspect.getfullargspec(constructor)[0]
    return len(arg_names) > 1 and arg_names[:2] == ["input_domain", "input_metric"]


def to_then(constructor) -> Callable[..., _PartialConstructor]:
    """Convert any `make_` constructor to a `then_` constructor"""

    if not supports_partial(constructor):  # pragma: no cover
        raise ValueError("the first two arguments of the constructor must be input_domain and input_metric")
    
    def then_func(*args, **kwargs):
        return _PartialConstructor(
            lambda input_domain, input_metric: constructor(
                input_domain, input_metric, *args, **kwargs
            )
        )

    then_func.__name__ = constructor.__name__.replace("make_", "then_", 1)
    then_func.__doc__ = f"""{constructor.__doc__}

    Fixes the ``input_domain`` and ``input_metric`` of: ``{constructor.__name__}``"""

    # preserve the signature, except for the first two parameters
    original_sig = inspect.signature(constructor)
    new_params = list(original_sig.parameters.values())[2:]
    then_func.__signature__ = original_sig.replace(parameters=new_params) # type: ignore[attr-defined]
    return then_func


def with_privacy(
    t_constructor: Callable[[Domain, Metric], Transformation],
) -> Callable[..., Union[Transformation, Measurement]]:
    from opendp.mod import assert_features
    from opendp.measurements import then_gaussian, then_laplace

    def private_constructor(input_domain, input_metric, privacy_measure, scale,
                            *args, **kwargs):
        assert_features("contrib")
        m_constructor = {
            "ZeroConcentratedDivergence": then_gaussian,
            "MaxDivergence": then_laplace,
        }[str(privacy_measure.type)]

        return (t_constructor(input_domain, input_metric, *args, **kwargs)
                >> m_constructor(scale)) # type: ignore[operator]

    private_constructor.__name__ = t_constructor.__name__.replace(
        "make_", "make_private_"
    )
    if t_constructor.__doc__ is not None:
        private_constructor.__doc__ = t_constructor.__doc__.replace(
            "Transformation", "Measurement"
        )
    return private_constructor



from opendp.extras._utilities import to_then
from opendp.extras.numpy import _sscp_domain
from opendp.mod import Domain, Metric, Transformation
from opendp._internal import _make_transformation

# planning to make this public, but may make more API changes


def make_np_sscp(input_domain: Domain, input_metric: Metric) -> Transformation:
    """Construct a Transformation that computes a covariance matrix from the input data.

    :param input_domain: instance of `array2_domain(size=_, num_columns=_)`
    :param input_metric: instance of `symmetric_distance()`

    :return: a Measurement that computes the DP sum
    """
    import opendp.prelude as dp

    dp.assert_features("contrib", "floating-point")

    if not str(input_domain).startswith("NPArray2Domain"):
        raise ValueError(
            f"input_domain ({input_domain}) must be a 2d-numpy array domain"
        )  # pragma: no cover

    if input_metric != dp.symmetric_distance():
        raise ValueError(f"input_metric ({input_metric}) must be symmetric distance")  # pragma: no cover
    
    input_desc = input_domain.descriptor

    if input_desc.nan:
        raise ValueError(f"input_domain ({input_domain}) must not contain NaN values")  # pragma: no cover

    if input_desc.num_columns is None:
        raise ValueError(f"input_domain ({input_domain}) must have known num_columns")  # pragma: no cover

    return _make_transformation(
        input_domain,
        input_metric,
        _sscp_domain(
            num_features=input_desc.num_columns,
            norm=input_desc.norm,
            p=input_desc.p,
            size=input_desc.size,
            T=input_desc.T,
        ),
        input_metric,
        lambda arg: arg.T @ arg,
        lambda d_in: d_in,
    )


# generate then variant of the constructor
then_np_sscp = to_then(make_np_sscp)



"""AIM mechanism from `MMSM22 <https://arxiv.org/abs/2201.12677>`_."""

from dataclasses import dataclass, replace
from itertools import combinations, chain
from math import pi, sqrt, prod

from opendp._lib import import_optional_dependency
from opendp.combinators import (
    make_adaptive_composition,
    make_fully_adaptive_composition,
    make_privacy_filter,
)
from opendp.domains import atom_domain, vector_domain
from opendp.extras.mbi._utilities import (
    TypedDictDomain,
    get_associated_metric,
    make_noise_marginal,
    make_stable_marginals,
    prior,
    weight_marginals,
    Count,
    Algorithm,
)
from opendp.measurements import then_noisy_max
from opendp.measures import max_divergence, zero_concentrated_divergence
from opendp.metrics import linf_distance
from opendp.mod import (
    ExtrinsicDomain,
    FrameDistance,
    LazyFrameDomain,
    Measure,
    Measurement,
    Metric,
    OdometerQueryable,
    Queryable,
    Transformation,
    binary_search_chain,
)
from opendp._internal import _make_transformation, _new_pure_function

from typing import Any, Optional, Sequence, TypeAlias, cast, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from opendp.extras.polars import Bound


@dataclass(kw_only=True, frozen=True)
class AIM(Algorithm):
    """AIM mechanism from `MMSM22 <https://arxiv.org/abs/2201.12677>`_.

    Adaptively chooses and estimates the least-well-approximated marginal.
    The stronger the correlation amongst a clique of columns, 
    the more likely AIM is to select the clique.

    The algorithm starts with a small per-step privacy budget, 
    and in each step increases the budget if the last measured marginal doesn't sufficiently improve the model.

    ..
        >>> import pytest  # `pip install opendp[mbi]` is necessary
        >>> _ = pytest.importorskip("mbi")

    .. code:: pycon

        >>> import opendp.prelude as dp
        >>> import polars as pl

        >>> dp.enable_features("contrib")

        >>> context = dp.Context.compositor(
        ...     data=pl.scan_csv(dp.examples.get_france_lfs_path(), ignore_errors=True),
        ...     privacy_unit=dp.unit_of(contributions=36),
        ...     privacy_loss=dp.loss_of(rho=0.1, delta=1e-7),
        ... )

        >>> table_aim = (
        ...     context.query(rho=0.1, delta=1e-7)
        ...     # transformations/truncation may be applied here
        ...     .select("SEX", "AGE", "HWUSUAL", "ILOSTAT")
        ...     .contingency_table(
        ...         keys={"SEX": [1, 2]}, 
        ...         cuts={"AGE": [20, 40, 60], "HWUSUAL": [1, 20, 40]},
        ...         algorithm=dp.mbi.AIM()
        ...     )
        ...     .release()
        ... )

        >>> table_aim.synthesize() # doctest: +SKIP
        shape: (3_807_732, 4)
        ┌─────┬───────────┬───────────┬─────────┐
        │ SEX ┆ AGE       ┆ HWUSUAL   ┆ ILOSTAT │
        │ --- ┆ ---       ┆ ---       ┆ ---     │
        │ i64 ┆ f64       ┆ f64       ┆ i64     │
        ╞═════╪═══════════╪═══════════╪═════════╡
        │ 1   ┆ 55.446336 ┆ 20.776579 ┆ 1       │
        │ 1   ┆ 28.21838  ┆ 40.53348  ┆ 1       │
        │ 2   ┆ 43.291215 ┆ 34.406155 ┆ 1       │
        │ 1   ┆ 55.106615 ┆ 22.413161 ┆ 1       │
        │ 2   ┆ 42.585227 ┆ 40.11279  ┆ 3       │
        │ …   ┆ …         ┆ …         ┆ …       │
        │ 1   ┆ 58.197292 ┆ 40.139579 ┆ 1       │
        │ 1   ┆ 59.371221 ┆ 19.671153 ┆ 1       │
        │ 2   ┆ 19.862917 ┆ 40.339046 ┆ 9       │
        │ 1   ┆ 19.492355 ┆ 32.233661 ┆ 1       │
        │ 2   ┆ 60.863244 ┆ 40.737908 ┆ 3       │
        └─────┴───────────┴───────────┴─────────┘

    The algorithm is similar to the Multiplicative Weights Exponential Mechanism (MWEM) introduced in `HLM10 <https://arxiv.org/abs/1012.4763>`_,
    in that the exponential mechanism selects a marginal in each step.

    AIM differs from MWEM in that the distribution is represented via a graphical model and fitted via mirror descent,
    instead of joint distribution densities and the multiplicative weights update rule.
    This allows AIM to support higher-dimensional datasets.
    """

    queries: list[Count] | int = 3
    """Explicit workload of interactions, or maximum degree of interactions to consider."""
    measure_split: float = 0.9
    """Remaining proportion of budget to allocate to measuring marginals.
    
    The complement is spent on selecting marginals."""
    max_size: float = 80.0
    """Maximum memory constraint in MB for the marginal selection."""
    rounds: Optional[int] = None
    """Maximum number of rounds to run the algorithm."""

    def __post_init__(self):
        super().__post_init__()

        if isinstance(self.queries, int) and self.queries < 1:
            raise ValueError(f"queries ({self.queries}) must be positive")

        if isinstance(self.queries, list) and not self.queries:
            raise ValueError("queries must not be non-empty")

        if not (0 < self.measure_split <= 1):
            raise ValueError(f"measure_split ({self.measure_split}) must be in (0, 1]")

        if self.max_size <= 0.0:
            raise ValueError(f"max_size ({self.max_size}) must be positive")

    def make_marginals(
        self,
        input_domain: LazyFrameDomain,
        input_metric: FrameDistance,
        output_measure: Measure,
        d_in: list["Bound"],
        d_out: float,
        *,
        marginals: dict[tuple[str, ...], Any],
        model,  # MarkovRandomField
    ) -> Measurement:
        """Implements AIM (Adaptive Iterative Mechanism) for ordinal data.

        :param input_domain: domain of input data
        :param input_metric: how to compute distance between datasets
        :param output_measure: how to measure privacy of release
        :param d_in: distance between adjacent input datasets
        :param d_out: upper bound on the privacy loss
        :param marginals: prior marginal releases
        :param model: warm-start fit of MarkovRandomField
        """
        import_optional_dependency("mbi")
        from mbi import MarkovRandomField  # type: ignore[import-untyped,import-not-found]

        if not isinstance(model, MarkovRandomField):
            raise ValueError("model must be a MarkovRandomField")

        queries = _expand_queries(self.queries, input_domain.columns)
        algorithm = replace(self, queries=queries)

        lp_metric = get_associated_metric(output_measure)
        cliques = [query.by for query in queries]

        t_marginals = make_stable_marginals(
            input_domain, input_metric, lp_metric, cliques
        )
        d_marginals = t_marginals.map(d_in)

        T = algorithm.rounds or (16 * len(input_domain.columns))

        def function(
            qbl: OdometerQueryable,
        ) -> tuple[dict[tuple[str, ...], Any], MarkovRandomField]:
            # mutable state
            current_model = model
            current_marginals = marginals.copy()
            d_step = d_out / T
            d_spent = 0.0

            while d_step:
                m_step = _make_aim_marginal(
                    *t_marginals.output_space,
                    output_measure,
                    t_marginals.map(d_in),
                    d_step,
                    marginals=current_marginals,
                    model=current_model,
                    max_size=algorithm.max_size * (d_spent + d_step) / d_out,
                    algorithm=algorithm,
                )
                if not m_step:
                    break

                R: TypeAlias = tuple[
                    dict[tuple[str, ...], Any], MarkovRandomField, bool
                ]
                current_marginals, current_model, is_significant = cast(R, qbl(m_step))

                if not is_significant:
                    d_step *= 4

                d_spent = qbl.privacy_loss(d_marginals)
                d_step = min(d_step, max(prior(d_out - d_spent), 0))

            return current_marginals, current_model

        return (
            t_marginals
            >> make_privacy_filter(
                make_fully_adaptive_composition(
                    *t_marginals.output_space, output_measure
                ),
                d_in=d_marginals,
                d_out=d_out,
            )
            >> _new_pure_function(function)
        )


def _make_aim_marginal(
    input_domain: ExtrinsicDomain,  # TypedDictDomain of {clique: d-dimensional counts}
    input_metric: Metric,
    output_measure: Measure,
    d_in: int,
    d_out: float,
    marginals: dict[tuple[str, ...], Any],
    model,  # MarkovRandomField
    max_size: float,
    algorithm: AIM,
) -> Optional[Measurement]:
    """Create an interactive measurement that computes one step of the AIM algorithm."""
    import numpy as np  # type: ignore[import-not-found]
    from mbi import MarkovRandomField, LinearMeasurement  # type: ignore[import-not-found]
    from opendp.extras.numpy import NPArrayDDomain

    model = cast(MarkovRandomField, model)

    cardinalities = {
        c: d.cast(NPArrayDDomain).shape
        for c, d in input_domain.cast(TypedDictDomain).items()
    }

    # factors to convert stddev -> scale and scale -> half-distribution expectation
    if output_measure == max_divergence():
        to_scale, to_mu = sqrt(2), 1.0
    elif output_measure == zero_concentrated_divergence():
        to_scale, to_mu = 1.0, sqrt(2 / pi)

    d_measure = d_out * algorithm.measure_split
    d_select = max(prior(prior(d_out - d_measure)), 0)

    m_select = _make_aim_select(  # type: ignore[misc]
        input_domain,
        input_metric,
        output_measure,
        queries=algorithm.queries,  # type: ignore[arg-type]
        model=model,
        d_in=d_in,
        d_out=d_select,
        max_size=max_size,
    )

    if not m_select:
        return None

    def function(
        qbl: Queryable,
    ) -> tuple[dict[tuple[str, ...], Any], MarkovRandomField, bool]:
        # SELECT a query that best reduces the error
        selected_clique = qbl(m_select)

        # MEASURE selected marginal with noise
        m_measure = binary_search_chain(
            lambda s: make_noise_marginal(
                input_domain,
                input_metric,
                output_measure,
                clique=selected_clique,
                scale=s,
            ),
            d_in=d_in,
            d_out=d_measure,
            T=float,
        )

        new_marginal: LinearMeasurement = qbl(m_measure)

        # GENERATE an updated probability distribution
        prev_tab = model.project(selected_clique).values

        all_marginals = weight_marginals(marginals, new_marginal)

        new_model: MarkovRandomField = algorithm.estimator(
            model.domain,
            list(all_marginals.values()),
            potentials=model.potentials.expand(list(all_marginals.keys())),
        )

        next_tab = new_model.project(selected_clique).values
        diff = float(np.linalg.norm((next_tab - prev_tab).flatten(), 1))

        size = prod(cardinalities[selected_clique])
        mean = new_marginal.stddev * to_scale * to_mu * size

        # Update is significant if the L1 norm of the update to the selected marginal
        #    is greater than the expectation of the noise's half-distribution.
        is_significant = diff > mean

        return all_marginals, new_model, is_significant

    return make_adaptive_composition(
        input_domain,
        input_metric,
        output_measure,
        d_in=d_in,
        d_mids=[d_select, d_measure],
    ) >> _new_pure_function(function)


def _make_aim_select(
    input_domain: ExtrinsicDomain,
    input_metric: Metric,
    output_measure: Measure,
    d_in,
    d_out,
    queries: list[Count],
    model,  # MarkovRandomField
    max_size: float,
) -> Optional[Measurement]:
    """Make a measurement that selects a set of marginal query that will minimize error."""
    import mbi

    # factor to convert scale -> half-distribution expectation
    if output_measure == max_divergence():
        to_mu = 1.0
    elif output_measure == zero_concentrated_divergence():
        to_mu = sqrt(2 / pi)

    model = cast(mbi.MarkovRandomField, model)

    def is_small(clique: tuple[str, ...]) -> bool:
        model_size = mbi.junction_tree.hypothetical_model_size(
            model.domain, cliques=model.cliques + [clique]
        )
        return model_size <= max_size

    # only keep queries that fit the memory constraint
    candidates = [q for q in queries if is_small(q.by)]

    if not candidates:
        return None

    def make(scale: float) -> Measurement:
        return _make_aim_scores(
            input_domain, input_metric, candidates, scale * to_mu, model
        ) >> then_noisy_max(output_measure=output_measure, scale=scale)

    try:
        return binary_search_chain(
            make, d_in=d_in, d_out=d_out, T=float
        ) >> _new_pure_function(lambda idx: candidates[idx].by)
    except ValueError:  # pragma: no cover
        # can fail in the unlikely scenario where
        #    d_out is so small that no scale is suitable
        return None


def _make_aim_scores(
    input_domain: ExtrinsicDomain,
    input_metric: Metric,
    queries: list[Count],
    expectation: float,
    model,  # MarkovRandomField
) -> Transformation:
    """Make a transformation that assigns a score representing how poorly each query is estimated."""
    from opendp.extras.numpy import NPArrayDDomain
    import numpy as np
    from mbi import MarkovRandomField  # type: ignore[import-not-found]

    model = cast(MarkovRandomField, model)

    for value_domain in input_domain.cast(TypedDictDomain).values():
        value_domain.cast(NPArrayDDomain)  # pragma: no cover

    def score_query(query: Count, exact: np.ndarray):
        penalty = expectation * prod(exact.shape)
        synth = model.project(query.by).values

        return (np.linalg.norm((exact - synth).flatten(), 1) - penalty) * query.weight

    return _make_transformation(
        input_domain=input_domain,
        input_metric=input_metric,
        output_domain=vector_domain(atom_domain(T="f64", nan=False)),
        output_metric=linf_distance(T="f64", monotonic=False),
        function=lambda exact_tabs: [
            score_query(query, exact_tabs[query.by]) for query in queries
        ],
        stability_map=lambda d_in: max(d_in[q.by] * q.weight for q in queries),
    )


def _expand_queries(queries: list[Count] | int, columns: list[str]) -> list[Count]:
    """Expand queries to include all subsets.

    :param queries: Explicit queries to shape workload, or maximum degree to preserve.
    :param columns: All columns in the data, for use when queries is a degree.
    """
    if isinstance(queries, int):
        queries = [
            Count(by) for by in combinations(columns, min(queries, len(columns)))
        ]

    groupings = set(chain.from_iterable(_powerset(q.by) for q in queries))

    def compute_weight(new_by: tuple[str, ...]) -> float:
        return sum(q.weight * len(set(new_by) & set(q.by)) for q in queries)

    return [Count(by, compute_weight(by)) for by in groupings if by]


def _powerset(x: Sequence):
    """returns all subsets"""
    return chain.from_iterable(combinations(x, r) for r in range(len(x) + 1))



# type: ignore
def make_vec(
    input_space: tuple[AtomDomain[T], AbsoluteDistance[Q]],
) -> Transformation[
    AtomDomain[T], VectorDomain[AtomDomain[T]], AbsoluteDistance[Q], LpDistance[P, Q]
]:
    input_domain, input_metric = input_space
    return Transformation.new(
        input_domain,
        VectorDomain.new(input_domain).with_size(1),
        lambda arg: [arg],
        input_metric,
        LpDistance.default(),
        lambda d_in: d_in,
    )



'''
This module requires extra installs: ``pip install 'opendp[numpy]'``

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.numpy``.    
'''

from __future__ import annotations
from typing import NamedTuple, Literal, Optional
from opendp.mod import Domain
from opendp.typing import RuntimeType, RuntimeTypeDescriptor, _ELEMENTARY_TYPES, _PRIMITIVE_TYPES
from opendp._lib import import_optional_dependency
from opendp._internal import _extrinsic_domain
import typing
from opendp.extras.numpy._make_np_clamp import make_np_clamp, then_np_clamp # noqa: F401


if typing.TYPE_CHECKING: # pragma: no cover
    import numpy # type: ignore[import-not-found]

def _check_norm_and_p(norm: float | None, p: int | None):
    """Checks that a scalar L`p` `norm` is well-defined"""
    if (norm is None) != (p is None):
        raise ValueError("norm and p must both be set")

    if norm is not None:
        if isinstance(norm, int):
            norm = float(norm)
        if not isinstance(norm, float):
            raise ValueError("norm must be float")  # pragma: no cover
        if norm < 0.0:
            raise ValueError("norm must be non-negative")  # pragma: no cover

    if p not in {None, 1, 2}:
        raise ValueError("p must be 1 or 2")  # pragma: no cover


def _check_nonnegative_int(v: int | None, name: str):
    if v is not None:
        if not isinstance(v, int):
            raise ValueError(f"{name} must be an integer")  # pragma: no cover
        if v < 0:
            raise ValueError(f"{name} must be non-negative")  # pragma: no cover


def _fmt_attrs(attrs: NamedTuple) -> str:
    return ", ".join(f"{k}={v}" for k, v in attrs._asdict().items() if v is not None)


def array2_domain(
    *,
    norm: float | None = None,
    p: Literal[1, 2, None] = None,
    origin=None,
    size: int | None = None,
    num_columns: int | None = None,
    nan: Optional[bool] = None,
    cardinalities: list[int] | numpy.ndarray | None = None,
    T: RuntimeTypeDescriptor | None = None,
) -> Domain:
    """Construct a Domain representing 2-dimensional numpy arrays.

    :param norm: each row in x is bounded by the norm
    :param p: designates L`p` norm
    :param origin: center of the norm region. Assumed to be at zero
    :param size: number of rows in data
    :param num_columns: number of columns in the data
    :param nan: whether NaN values are allowed
    :param cardinalities: cardinalities of the categorical columns
    :param T: atom type
    """
    np = import_optional_dependency('numpy')
    import opendp.prelude as dp

    _check_norm_and_p(norm, p)

    if norm is not None:
        # normalize origin to a scalar
        origin = origin if origin is not None else 0.0

    if norm is None and origin is not None:
        raise ValueError("origin may only be set if data has bounded norm")  # pragma: no cover

    if isinstance(origin, (int, float)):
        # normalize origin to a 1d-ndarray
        origin = np.array(origin)

    if isinstance(origin, np.ndarray):
        if origin.dtype.kind in {"i", "u"}:
            origin = origin.astype(float)

        if origin.dtype.kind != "f":
            raise ValueError("origin array must be numeric")

        if origin.ndim == 0:
            if origin != 0:
                raise ValueError("scalar origin must be zero")
            if num_columns is not None:
                # normalize to a 1d-ndarray
                origin = np.repeat(origin, num_columns)

        if origin.ndim == 1:
            if num_columns is None:
                num_columns = origin.size
            if num_columns != origin.size:
                raise ValueError(f"origin must have num_columns={num_columns} values")

        if origin.ndim not in {0, 1}:
            raise ValueError("origin must have 0 or 1 dimensions")  # pragma: no cover

    elif origin is not None:
        raise ValueError("origin must be a scalar, ndarray or None")

    _check_nonnegative_int(size, "size")
    _check_nonnegative_int(num_columns, "num_columns")

    if isinstance(cardinalities, list):
        cardinalities = np.asarray(cardinalities)

    if isinstance(cardinalities, np.ndarray):
        if cardinalities.ndim != 1:
            raise ValueError(f"cardinalities ndim ({cardinalities.ndim}) must be one")
        
        if not np.issubdtype(cardinalities.dtype, np.integer):
            raise ValueError(f"cardinalities dtype ({cardinalities.dtype}) must be integer")
        
        if any(c <= 0 for c in cardinalities):
            raise ValueError(f"cardinalities ({cardinalities}) must be positive")
        
        if num_columns is None:
            num_columns = len(cardinalities)

        if len(cardinalities) != num_columns:
            raise ValueError(f"cardinalities length ({len(cardinalities)}) must match num_columns ({num_columns})")
    
    elif cardinalities is not None:
        raise ValueError("cardinalities must be a list, ndarray or None")
    
    T = T or _ELEMENTARY_TYPES.get(origin.dtype.type)
    if T is None:
        raise ValueError("must specify T, the type of data in the array")  # pragma: no cover
    T = dp.RuntimeType.parse(T)
    if T not in _PRIMITIVE_TYPES:
        raise ValueError(f"T ({T}) must be a primitive type")

    if nan is None:
        nan = T in {"f32", "f64"}

    def _member(x):
        if not isinstance(x, np.ndarray):
            raise TypeError("must be a numpy ndarray")
        T_actual = _ELEMENTARY_TYPES.get(x.dtype.type)
        if T_actual != T:
            raise TypeError(f"must have data of type {T}, got {T_actual}")
        if x.ndim != 2:
            raise ValueError("must be a 2-dimensional array")
        if num_columns is not None and x.shape[1] != num_columns:
            raise ValueError(f"must have {num_columns} columns")
        
        if T in {"f32", "f64"} and not nan and np.isnan(x).any():
            raise ValueError("must not contain NaN values")

        if origin is not None:
            x = x - origin
        if norm is not None and np.linalg.norm(x, ord=p, axis=1).max() > norm:
            raise ValueError(f"must have row norm at most {norm}")
        if size is not None and len(x) != size:
            raise ValueError(f"must have exactly {size} rows")
        
        if cardinalities is not None:
            n_unique = np.array([len(np.unique(x_i)) for x_i in x.T])
            if any(cardinalities < n_unique):
                msg = f"unique values in data ({n_unique}) must not exceed cardinalities ({cardinalities})"
                raise ValueError(msg)
        
        return True

    desc = NPArray2Domain(
        origin=origin,
        norm=norm,
        p=p,
        size=size,
        num_columns=num_columns,
        nan=nan,
        cardinalities=cardinalities,
        T=T,
    )

    return _extrinsic_domain(f"NPArray2Domain({_fmt_attrs(desc)})", _member, desc)


class NPArray2Domain(NamedTuple):
    origin: numpy.ndarray | None
    norm: float | None
    p: Literal[1, 2, None]
    size: int | None
    num_columns: int | None
    nan: bool
    cardinalities: numpy.ndarray | None
    T: str | RuntimeType

# Without these we get: "Alias for field number ..."
NPArray2Domain.origin.__doc__ = 'center of the norm region'
NPArray2Domain.norm.__doc__ = 'each row in x is bounded by the norm'
NPArray2Domain.p.__doc__ = 'designates L`p` norm'
NPArray2Domain.size.__doc__ = 'number of rows in data'
NPArray2Domain.num_columns.__doc__ = 'number of columns in the data'
NPArray2Domain.nan.__doc__ = 'whether NaN values are allowed'
NPArray2Domain.cardinalities.__doc__ = 'cardinalities of the categorical columns'
NPArray2Domain.T.__doc__ = 'atom type'


def arrayd_domain(
    *,
    shape: tuple[int, ...],
    T: RuntimeTypeDescriptor,
) -> Domain:
    """Construct a Domain representing d-dimensional numpy arrays.

    :param shape: shape of the array
    :param T: atom type
    """
    np = import_optional_dependency('numpy')
    import opendp.prelude as dp

    if not isinstance(shape, tuple):
        raise ValueError(f"shape ({shape}) must be a tuple")
    if any(not isinstance(s, int) or s <= 0 for s in shape):
        raise ValueError(f"shape ({shape}) must be a tuple of positive integers")

    if T is None:
        raise ValueError("must specify T, the type of data in the array")  # pragma: no cover
    T = dp.RuntimeType.parse(T)
    if T not in _PRIMITIVE_TYPES:
        raise ValueError(f"T ({T}) must be a primitive type")

    def _member(x):
        if not isinstance(x, np.ndarray):
            raise ValueError("must be a numpy ndarray")
        T_actual = _ELEMENTARY_TYPES.get(x.dtype.type)
        if T_actual != T:
            raise ValueError(f"must have data of type {T}, got {T_actual}")
        
        if x.shape != shape:
            raise ValueError(f"must have shape {shape}")
        return True

    desc = NPArrayDDomain(
        shape=shape,
        T=T,
    )

    return _extrinsic_domain(f"NPArrayDDomain({_fmt_attrs(desc)})", _member, desc)


class NPArrayDDomain(NamedTuple):
    shape: tuple[int, ...]
    T: str | RuntimeType


NPArrayDDomain.shape.__doc__ = 'shape of the array'
NPArrayDDomain.T.__doc__ = 'atom type'


def _sscp_domain(
    *,
    norm: float | None = None,
    p: Literal[1, 2, None] = None,
    size: int | None = None,
    num_features: int | None = None,
    T: RuntimeTypeDescriptor = float,
) -> Domain:
    """The domain of sums of squares and cross products matrices formed by computing x^Tx,
    for some dataset x.

    Elements are finite, members are square symmetric positive semi-definite matrices.

    :param norm: each row in x is bounded by the norm
    :param p: designates L`p` norm
    :param size: number of rows in data
    :param num_features: number of rows/columns in the matrix
    :param T: the type of data elements in the array
    """
    import opendp.prelude as dp
    np = import_optional_dependency('numpy')

    _check_norm_and_p(norm, p)
    _check_nonnegative_int(size, "size")
    _check_nonnegative_int(num_features, "num_features")

    T = dp.RuntimeType.parse(T)
    if T not in {dp.f32, dp.f64}:
        raise ValueError("T must be a float type")

    def _member(x):
        if not isinstance(x, np.ndarray):
            raise TypeError("must be a numpy ndarray")
        T_actual = _ELEMENTARY_TYPES.get(x.dtype.type)
        if T_actual != T:
            raise TypeError(f"must have data of type {T}, got {T_actual}")
        if x.shape != (x.shape[0], x.shape[0]):
            raise ValueError("must be a square array")
        if num_features is not None and x.shape[0] != num_features:
            raise ValueError(f"must have {num_features} features")
        if (~np.isfinite(x)).any():
            raise ValueError("must have finite values")
        if np.any(x != x.T):
            raise ValueError("must be symmetric")
        if np.any(np.linalg.eigvals(x) < 0):
            raise ValueError("must be positive semi-definite")
        return True

    class NPSSCPDescriptor(NamedTuple):
        num_features: int | None
        norm: float | None
        p: Literal[1, 2, None]
        size: int | None
        T: str | dp.RuntimeType

    desc = NPSSCPDescriptor(
        num_features=num_features,
        norm=norm,
        p=p,
        size=size,
        T=T,
    )

    return _extrinsic_domain(f"NPSSCPDomain({_fmt_attrs(desc)})", _member, desc)



# type: ignore
def make_noisy_top_k(
    input_domain: VectorDomain[AtomDomain[TIA]],
    input_metric: LInfDistance[TIA],
    privacy_measure: MO,
    k: usize,
    scale: f64,
    negate: bool,
) -> Measurement:
    if input_domain.element_domain.nan():  # |\label{check-non-nan}|
        raise "input domain elements must be non-nan"

    if input_domain.size is not None:
        if k > input_domain.size:
            raise "k must not exceed the number of candidates"

    if not scale.is_finite() or scale.is_sign_negative():  # |\label{check-non-negative-scale}|
        raise "scale must be finite and non-negative"

    monotonic = input_metric.monotonic

    def privacy_map(d_in: TIA):  # |\label{fn-privacy-map}|
        # convert to range distance
        d_in = d_in if monotonic else d_in.inf_add(d_in)
        d_in = f64.inf_cast(d_in)  # |\label{fn-inf-cast}|

        if d_in.is_sign_negative():  # |\label{din-non-neg}|
            raise "sensitivity must be non-negative"

        if d_in.is_zero():  # |\label{din-non-zero}|
            return 0.0

        if scale.is_zero():  # |\label{scale-zero}|
            return f64.INFINITY

        # |\label{fn-privacy-map-call}|
        return MO.privacy_map(d_in, scale).inf_mul(f64.inf_cast(k))

    return Measurement.new(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=privacy_measure,
        function=lambda x: MO.noisy_top_k(x, scale, k, negate),
        privacy_map=privacy_map,
    )



# type: ignore
def make_privacy_filter(
    odometer: Odometer[DI, MI, MO, Q, A],
    d_in: MI_Distance,
    d_out: MO_Distance,
) -> Measurement[DI, MI, MO, OdometerQueryable[MI, MO, Q, A]]:
    odo_function = odometer.function
    continuation_rule = new_continuation_rule(  # `\label{continuation-rule}`
        d_in, d_out, Q, A, MI_Distance, MO_Distance
    )

    def function(arg: DI_Carrier) -> OdometerQueryable[MI, MO, Q, A]:
        return wrap(continuation_rule, lambda: odo_function.eval(arg))  # `\label{eval}`

    def privacy_map(d_in_p: MI_Distance) -> MO_Distance:
        if d_in_p.total_gt(d_in):
            raise "input distance must not be greater than d_in"

        return d_out

    return Measurement.new(
        odometer.input_domain,
        odometer.input_metric,
        odometer.output_measure,
        Function.new_interactive(function),
        PrivacyMap.new_fallible(privacy_map),
    )



# type: ignore
def make_stable_group_by(
    input_domain: DslPlanDomain, input_metric: FrameDistance[M], plan: DslPlan
) -> Transformation[DslPlanDomain, DslPlanDomain, FrameDistance[M], FrameDistance[M]]:
    match plan:
        case DslPlan.GroupBy(input, keys, aggs, apply, maintain_order, options):
            pass
        case _:
            raise "Expected group by in logical plan"

    if apply is not None:
        raise "apply is not currently supported"

    if maintain_order:
        raise "maintain_order is wasted compute because row ordering is protected information"

    if options != GroupbyOptions.default():
        raise "options is not currently supported"

    t_prior = input.make_stable(input_domain, input_metric)
    middle_domain, middle_metric = t_prior.output_space()

    expr_domain = WildExprDomain(
        columns=middle_domain.series_domains,
        context=Context.RowByRow,
    )

    for key in keys:  # |\label{line:stable-keys}|
        key.make_stable(expr_domain, PartitionDistance(middle_metric[0]))

    for agg in aggs:  # |\label{line:infallible-aggs}|
        check_infallible(agg, Resize.Allow)

    if middle_metric[0].identifier().is_some():  # |\label{line:identifier}|
        raise "identifier is not currently supported"

    # prepare output domain series
    output_schema = middle_domain.simulate_schema(  # |\label{line:simulate-schema}|
        lambda lf: lf.group_by(keys).agg(aggs)
    )
    series_domains = [SeriesDomain.new_from_field(f) for f in output_schema]

    # prepare output domain margins
    h_keys = list(keys)

    def without_invariant(m: Margin) -> Margin:
        m.invariant = None
        return m

    margins = [
        without_invariant(m) for m in middle_domain.margins if m.by.is_subset(h_keys)
    ]

    output_domain = FrameDomain.new_with_margins(series_domains, margins)

    def stability_map(d_in: Bounds) -> Bounds:
        # |\label{line:total-contributions}|
        contributed_rows = d_in.get_bound(set()).per_group
        # |\label{line:total-groups}|
        contributed_groups = d_in.get_bound(h_keys).num_groups

        influenced_groups = option_min(contributed_rows, contributed_groups)
        if influenced_groups.is_none():
            return "an upper bound on the number of contributed rows or groups is required"

        if per_group is not None: # |\label{line:double}|
            per_group = per_group.inf_mul(2)

        bound = Bound(by=set(), per_group=per_group, num_groups=None)
        return Bounds([bound]) # |\label{line:bound}|

    t_group_agg = Transformation.new(
        middle_domain,
        output_domain,
        lambda plan: DslPlan.GroupBy(
            input=plan,
            keys=keys,
            aggs=aggs,
            apply=None,
            maintain_order=False,
            options=options,
        ),
        middle_metric,
        middle_metric,
        StabilityMap.new_fallible(stability_map),
    )

    return t_prior >> t_group_agg



# type: ignore
from math import prod

def get_margin(domain: FrameDomain, by: set[Expr]) -> Margin:
    margin = next(  # |\label{let-margin}|
        (m for m in domain.margins if m.by == by), 
        Margin.by(by)
    )

    subset_margins = [  # |\label{let-subset-margins}|
        margin
        for margin in domain.margins
        if margin.by.issubset(by)
    ]

    margin.max_length = min(  # |\label{let-max-length}|
        m.max_length
        for m in subset_margins
        if m.max_length is not None
    )

    all_max_groups = [  # |\label{all-max-groups}|
        (m.by, m.max_groups)
        for m in domain.margins
        if m.max_groups is not None
    ]
    max_groups_covering = find_min_covering(grouping_columns, all_max_groups)
    if max_groups_covering is not None:
        margin.max_groups = prod(v for _, v in max_groups_covering)

    all_invariants = (  # |\label{all-invariants}|
        m.invariant
        for m in domain.margins
        if by.issubset(m.by)
    )
    margin.invariant = max(all_invariants, key={None: 0, "Keys": 1, "Lengths": 2}.get)

    return margin



# type: ignore
def permute_and_flip(x: list[RBig], scale: RBig):
    if scale.is_zero():  # `\label{zero-scale}`
        return max(range(x.len()), key=lambda i: x[i])

    # begin nonzero scale `\label{nonzero-scale}`
    x_max = max(x)
    permutation = list(range(x.len()))

    for left in range(x.len()):
        right = left + sample_uniform_uint_below(x.len() - left)
        permutation.swap(left, right) # fisher-yates shuffle up to left `\label{shuffle}`

        candidate = permutation[left]
        if sample_bernoulli_exp((x_max - x[candidate]) / scale):
            return candidate
    
    raise "at least one x[candidate] is equal to x_max"


"""
This module requires extra installs: ``pip install 'opendp[scikit-learn]'``

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.sklearn.linear_model``.

If you're interested in the underlying algorithm, we've also
`implemented Theil-Sen Regression as a demonstration of OpenDP plugins <../user-guide/plugins/theil-sen-regression.html>`_.
"""

from typing import Iterable
from opendp.extras.sklearn.linear_model._make_private_theil_sen import (
    make_private_theil_sen as _make_private_theil_sen,
)  # noqa: F401
from opendp._lib import import_optional_dependency
from opendp.mod import Measure

__all__ = ["LinearRegression"]


class LinearRegression:
    """
    DP Linear Regression

    The interface is parallel to that offered by sklearn's
    `LinearRegression <https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html>`_.
    The ``fit`` method returns an sklearn ``LinearRegression`` object.

    :param x_bounds: Bounds for training data; For the moment, only lists containing a single tuple are supported
    :param y_bounds: Bounds for target data
    :param scale: The scale of the noise to be added
    :param runs: Controls how many times randomized pairwise predictions are computed. Increasing this value can improve the robustness and accuracy of the results; However, it can also increase computational cost and amount of noise needed later in the algorithm.
    :param candidates_count: How many evenly spaced candidates to generate
    :param fraction_bounds: predict y values at these cut percentiles of x_bounds.
    """

    def __init__(
        self,
        output_measure: Measure,
        x_bounds: Iterable[tuple[float, float]],
        y_bounds: tuple[float, float],
        scale: float,
        runs: int = 1,
        candidates_count: int = 100,
        fraction_bounds: tuple[float, float] = (0.25, 0.75),
    ):
        x_bounds = list(x_bounds)  # Shouldn't be so large that this is a problem
        if len(x_bounds) != 1:
            msg = f"For now, the x_bounds array must consist of a single tuple, not {x_bounds}"
            raise Exception(msg)
        
        self.measurement = _make_private_theil_sen(
            output_measure=output_measure,
            x_bounds=x_bounds[0],
            y_bounds=y_bounds,
            scale=scale,
            runs=runs,
            candidates_count=candidates_count,
            fraction_bounds=fraction_bounds,
        )

    def fit(
        self,
        X,
        y,
    ):
        """
        Fit DP linear model.

        :param X: Training data. Array-like of shape (n_samples, 1)
        :param y: Target values. Array-like of shape (n_samples,)
        :return: A fitted sklearn ``LinearRegression``

        :example:

        >>> import opendp.prelude as dp
        >>> try:
        ...    import sklearn
        ... except ModuleNotFoundError:
        ...     import pytest
        ...     pytest.skip('Requires extra install')
        >>> dp.enable_features("floating-point")
        >>> lin_reg = dp.sklearn.linear_model.LinearRegression(
        ...     dp.max_divergence(),
        ...     x_bounds=[(0, 10)],
        ...     y_bounds=(0, 10),
        ...     scale=1,
        ... ).fit(
        ...     X=[[1], [2], [3], [4], [5]],
        ...     y=[1, 2, 3, 4, 5],
        ... )
        >>> lin_reg.predict([[10]])
        array([...])
        """
        np = import_optional_dependency("numpy")
        from sklearn.linear_model import LinearRegression as LR

        X = np.array(X)
        slope, intercept = self.measurement(np.stack([X[:, 0], y], axis=1))

        fit_regression = LR()
        fit_regression.coef_ = np.array([slope])
        fit_regression.intercept_ = intercept
        return fit_regression

    def predict(X):
        """
        The ``fit()`` method returns a new sklearn object, so this method is never actually used.
        The sklearn documentation of the method with the same name is copied here only for reference.

        > Predict using the linear model.
        >
        > ### Parameters
        > *X : array-like or sparse matrix, shape (n_samples, n_features)*
        >
        > Samples.
        >
        > ### Returns
        > *C : array, shape (n_samples,)*
        >
        > Returns predicted values.

        .. end-markdown

        :raises NotImplementedError: This method is included only for documention.
        """
        raise NotImplementedError("Included only for documentation")  # pragma: no cover



from pathlib import Path
from json import loads
import pytest
import re
from opendp import measurements, transformations

@pytest.mark.parametrize(
    "module,function",
    [
        (m, f)
        for m in [measurements, transformations]
        for f in [
            getattr(m, f_name)
            for f_name in dir(m)
            if f_name.startswith('then_')
        ]
    ])
def test_thens_are_documented(module, function):
    m_name = module.__name__
    then_name = function.__name__
    make_name = then_name.replace('then_', 'make_')

    assert function.__doc__ is not None, 'missing documentation'
    assert f':py:func:`{m_name}.{make_name}`' in function.__doc__, f'no link to {make_name}'


docs_source = Path(__file__).parent.parent.parent / 'docs' / 'source'
assert docs_source.exists()


def get_self_and_parent(path: Path):
    return f'{path.parent.name}/{path.name}'

@pytest.mark.parametrize(
    "rst_path",
    list(docs_source.glob("**/*.rst")),
    ids=get_self_and_parent
)
def test_code_block_language(rst_path: Path):
    rst_lines = rst_path.read_text().splitlines()
    expected = ['pycon', 'rust', 'latex', 'r', 'shell']

    # For autoformatting of doctests, code blocks should be "pycon", not "python".
    # https://github.com/adamchainz/blacken-docs?tab=readme-ov-file#restructuredtext
    assert 'python' not in expected

    errors = []
    for i, line in enumerate(rst_lines):
        m = re.search(r'\.\.\s+code::\s+(\w+)', line)
        if m:
            language = m.group(1)
            if language not in expected:
                errors.append(f'line {i+1}: Got "{language}", expected one of: {", ".join(expected)}')
    assert not errors, '\n'.join(errors)


@pytest.mark.parametrize(
    "rst_path",
    list(docs_source.glob("**/*.rst")),
    ids=get_self_and_parent
)
def test_single_backticks(rst_path: Path):
    rst_lines = rst_path.read_text().splitlines()
    errors = []
    for i, line in enumerate(rst_lines):
        line = line.strip()
        if line.startswith(('>>>', '...', '//')):
            # Probably in a comment in a code sample: Skip.
            continue
        m = re.search(r'''
            ([^`:]|^)   # Non-backtick or start of line
            `           # backtick
            ([^`<>_:]+) # content, excluding RST links and tags
            `           # backtick
            ([^`]|$)    # Non-backtick or end of line
        ''', line, re.VERBOSE)
        if m:
            content = m.group(2)
            errors.append(f'line {i+1}: "{content}" will be italicized: add double-backticks, or change to "*".')
    assert not errors, '\n'.join(errors)


python_src = Path(__file__).parent.parent / 'src'
assert python_src.exists()


@pytest.mark.parametrize(
    "path",
    list(docs_source.glob("**/*.rst")) + list(python_src.glob("**/*.py")),
    ids=get_self_and_parent
)
def test_doctest_empty_lines(path: Path):
    lines = path.read_text().splitlines()
    errors = []
    for i, line in enumerate(lines):
        if re.search(r'>>>\s*$', line):
            errors.append(f'line {i+1}: wrap with "code::" block and drop empty ">>>".')
    assert not errors, '\n'.join(errors)


@pytest.mark.parametrize(
    "nb_path",
    list(docs_source.glob("**/*.ipynb")),
    ids=lambda path: path.name
)
def test_notebooks_are_executed(nb_path):
    nb = loads(nb_path.read_text())
    counts_sources = [(cell.get('execution_count'), ''.join(cell.get('source', ''))) for cell in nb['cells'] if cell['cell_type'] == 'code']
    triples = [(index, count, source) for (index, (count, source)) in enumerate(counts_sources, start=1)]
    indexes, counts, sources = zip(*triples)

    # Info for error message:
    bad_sources = [source for (index, count, source) in triples if index != count]
    from sys import version_info
    short_path = nb_path.relative_to(Path.cwd(), walk_up=True) if version_info >= (3, 12) else nb_path.name

    # Notebook execution requires dependencies (jupyter, matplotlib, ...) beyond the basic dev environment.
    assert indexes == counts, f'''Notebook not completely executed.
To fix: jupyter nbconvert --to notebook --execute {short_path} --inplace
First cell with missing or misordered execution:\n{bad_sources[0].splitlines()[0]}'''


def is_public_and_not_top(path: Path):
    '''
    >>> is_public_and_not_top(Path('_parent/child.py'))
    False
    >>> is_public_and_not_top(Path('_parent/__init__.py'))
    False
    >>> is_public_and_not_top(Path('parent/__init__.py'))
    True
    >>> is_public_and_not_top(Path('parent/_child.py'))
    False
    >>> is_public_and_not_top(Path('parent/child.py'))
    True
    '''
    return not path.parent.name.startswith('_') and (
        not path.name.startswith('_') or path.name == '__init__.py'
    ) and not (path.parent.name == 'extras' and path.name == '__init__.py')


public_extras = [
    path for path in 
    Path(__file__).parent.parent.glob("src/opendp/extras/**/*.py")
    if is_public_and_not_top(path)
]


def get_namespace(py_path: Path):
    # Getting __doc__ from the module might be cleaner,
    # but walking through the module hierarchy is trickier than path glob.
    ns = []
    parent_parent_name = py_path.parent.parent.name
    parent_name = py_path.parent.name
    stem = py_path.stem
    if parent_parent_name != 'extras' and parent_name != 'extras':
        ns.append(parent_parent_name)
    if parent_name != 'extras':
        ns.append(parent_name)
    if stem != '__init__':
        ns.append(stem)
    return f'dp.{".".join(ns)}'


def get_docstring(py_path: Path):
    import ast
    py = py_path.read_text()
    module_ast = ast.parse(py)
    # This is a little fragile, but as long as a docstring is first, it should work.
    return module_ast.body[0].value.value.strip() # type: ignore[attr-defined]
 

@pytest.mark.parametrize(
    "py_path",
    public_extras,
    ids=get_namespace
)
def test_extras_docstring_convenience(py_path):
    actual = get_docstring(py_path)
    expected_ns = get_namespace(py_path)
    expected = f"""
For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``{expected_ns}``.    
""".strip()
    assert expected in actual, f"expected not in actual: {expected=}\n{actual=}"


@pytest.mark.parametrize(
    "py_path",
    public_extras,
    ids=get_namespace
)
def test_extras_docstring_installs(py_path):
    actual = get_docstring(py_path)
    namespace = get_namespace(py_path)
    if 'sklearn' in namespace:
        extra = 'scikit-learn'
    elif 'numpy' in namespace:
        extra = 'numpy'
    elif 'polars' in namespace:
        extra = 'polars'
    elif 'mbi' in namespace:
        extra = 'mbi'
    elif 'examples' in namespace:
        pytest.skip("dp.examples does not need extra installs")
    expected = f"""
This module requires extra installs: ``pip install 'opendp[{extra}]'``   
""".strip()
    assert expected in actual, f"expected not in actual: {expected=}\n{actual=}"


# type: ignore
def sample_uniform_uint_below(upper: T) -> T:
    threshold = T.MAX - T.MAX % upper

    while True:
        sample = sample_from_uniform_bytes()
        if sample < threshold:
            return sample % upper



# type: ignore
def make_quantile_score_candidates(
    input_domain: VectorDomain[AtomDomain[TIA]],
    input_metric: MI,
    candidates: list[TIA],
    alpha: f64,
) -> Transformation:
    if input_domain.element_domain.nan():
        raise ValueError("input_domain members must have non-nan elements")

    check_candidates(candidates)

    size = input_domain.size
    if size is not None:
        size = u64.exact_int_cast(input_domain.size)

    alpha_num, alpha_den, size_limit = score_candidates_constants(size, alpha)

    def function(arg: list[TIA]) -> list[u64]:
        scores = compute_score(arg, candidates, alpha_num, alpha_den, size_limit)
        return Vec.from_iter(scores)  # like calling list(s) on an iter s

    return Transformation(
        input_domain=input_domain,
        output_domain=VectorDomain(
            element_domain=AtomDomain(T=u64), size=len(candidates)
        ),
        function=function,
        input_metric=input_metric,
        output_metric=LInfDistance.default(T=u64),
        stability_map=StabilityMap.new_fallible( # `\label{map}`
            score_candidates_map(
                alpha_num,
                alpha_den,
                input_domain.size.is_some(),
            )
        ),
    )



# type: ignore
def sample_discrete_laplace(scale) -> int:
    if scale == 0:
        return 0
        
    inv_scale = recip(scale)
    
    while True:
        sign = sample_standard_bernoulli()
        magnitude = sample_geometric_exp_fast(inv_scale) # |\label{line:magnitude}|
        
        if sign or magnitude != 0: # |\label{line:branching}|
            if sign:
                return magnitude
            else:
                return -magnitude


import opendp  # So we can access the private _PRIMITIVE_TYPES
from opendp._lib import import_optional_dependency
from opendp.extras.polars import Margin
import opendp.prelude as dp
import pytest


@pytest.mark.parametrize("ty", opendp.typing._PRIMITIVE_TYPES)
def test_atom_domain_primitive_types(ty):
    # Checks that all primitive types are construct-able over FFI.
    # This ensures that all Polars dtypes can be used in debug builds.
    dp.option_domain(dp.atom_domain(T=ty))


def test_atom_domain_bounds():
    atom_domain = dp.atom_domain(T=int, bounds=(1, 2))
    assert str(atom_domain) == "AtomDomain(bounds=[1, 2], T=i32)"
    assert atom_domain.carrier_type == dp.i32
    assert atom_domain.bounds == (1, 2)
    assert atom_domain != str(atom_domain)
    assert not atom_domain.nan

def test_atom_domain_nullable():
    atom_domain = dp.atom_domain(T=float, nan=True)
    assert atom_domain.carrier_type == dp.f64
    assert atom_domain.bounds is None
    assert atom_domain != str(atom_domain)
    assert atom_domain.nan


def test_option_domain():
    atom_domain = dp.atom_domain(bounds=(1.0, 2.0), nan=True)
    option_domain = dp.option_domain(atom_domain)
    assert option_domain.element_domain == atom_domain


def test_series_domain():
    atom_domain = dp.atom_domain(bounds=(1.0, 2.0), nan=True)
    series_domain = dp.series_domain("A", atom_domain)
    assert series_domain.name == "A"
    assert series_domain.element_domain == atom_domain
    assert not series_domain.nullable


def test_lazyframe_domain_series():
    atom_domain = dp.atom_domain(bounds=(1.0, 2.0), nan=True)
    series_domain = dp.series_domain("A", atom_domain)
    frame_domain = dp.lazyframe_domain([series_domain])

    assert frame_domain.get_series_domain("A") == series_domain
    assert frame_domain.columns == ["A"]


def test_lazyframe_domain_margins():
    import_optional_dependency("polars")
    atom_domain = dp.atom_domain(bounds=(1.0, 2.0), nan=True)
    series_domain_a = dp.series_domain("A", atom_domain)
    series_domain_b = dp.series_domain("B", atom_domain)
    frame_domain = dp.with_margin(
        dp.lazyframe_domain([series_domain_a, series_domain_b]),
        dp.polars.Margin(
            by=["A"],
            max_groups=20,
            max_length=1000,
            invariant="keys"
        ),
    )
    assert frame_domain.get_series_domain("A") == series_domain_a

    # for coverage
    assert dp.polars.Margin(by=[]) != "not a margin"
    assert dp.polars.Margin(by=[]) != dp.polars.Margin(by=["A"])
    assert frame_domain.get_margin([]) == dp.polars.Margin(
        by=[], max_groups=1, invariant="keys"
    )

    assert frame_domain.get_margin(["A"]) == dp.polars.Margin(
        by=["A"],
        max_groups=20,
        max_length=1000,
        invariant="keys",
    )

    assert frame_domain.get_margin(["A", "B"]) == dp.polars.Margin(
        by=["A", "B"],
        max_length=1000,
    )

    # now add a margin for column B
    frame_domain = dp.with_margin(
        frame_domain,
        Margin(
            by=["B"],
            max_groups=20,
            max_length=500,
            invariant="keys"
        ),
    )

    assert frame_domain.get_margin(["A", "B"]) == dp.polars.Margin(
        by=["A", "B"],
        max_length=500,  # from B
        max_groups=400, # 20 * 20
    )

    with pytest.raises(ValueError, match="must be a sequence type; Did you mean [\"A\"]?"):
        frame_domain.get_margin("A")

    with pytest.raises(ValueError, match="must be a sequence type"):
        frame_domain.get_margin(2) # type: ignore[arg-type]


def test_vector_domain():
    atom_domain = dp.atom_domain(bounds=(1.0, 2.0), nan=True)
    vector_domain = dp.vector_domain(atom_domain, size=10)
    assert vector_domain.size == 10
    assert vector_domain.element_domain == atom_domain



# type: ignore
class CompositionMeasure(ApproximateZeroConcentratedDivergence):
    def composability(  # |\label{line:composability}|
        self, adaptivity: Adaptivity
    ) -> Composability:
        return Composability.Sequential

    def compose(self, d_mids: Vec[Self_Distance]) -> Self_Distance:
        rho_g, del_g = 0.0, 0.0
        for rho_i, del_i in d_mids:
            rho_g = rho_g.inf_add(rho_i)
            del_g = del_g.inf_add(del_i)
        return rho_g, del_g



# type: ignore
def make_fully_adaptive_composition(
    input_domain: DI,
    input_metric: MI,
    output_measure: MO,
) -> Odometer[DI, MI, MO, Measurement[DI, MI, MO, TO], TO]:

    # check if fully adaptive composition is supported
    output_measure.composability(Adaptivity.FullyAdaptive)

    def function(
        arg: DI_Carrier,
    ) -> OdometerQueryable[Measurement[DI, MI, MO, TO], TO, MO_Distance]:
        return new_fully_adaptive_composition_queryable(
            input_domain, input_metric, output_measure, arg
        )

    return Odometer.new(
        input_domain, input_metric, output_measure, Function.new_fallible(function)
    )



'''
For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.examples``.    
'''

from pathlib import Path


# pragma: no cover to make local testing easier
def _http_get(url: str) -> bytes: # pragma: no cover
    '''
    Normally would use the requests library for this, but we want to avoid extra dependencies.

    This only does what we need. Does not handle:
    - plain http
    - redirects
    - url queries
    - explicit port numbers
    '''
    from urllib.parse import urlparse
    import http.client
    parsed = urlparse(url)
    assert parsed.hostname is not None
    conn = http.client.HTTPSConnection(parsed.hostname)
    conn.request("GET", parsed.path)
    response = conn.getresponse()

    assert response.status == 200, f"{response.status} {response.reason}\n{response.headers}"
    body = response.read()
    conn.close()
    return body


def get_california_pums_path() -> Path:
    '''
    Returns the path to a CSV derived from a
    PUMS (Public Use Microdata Sample) file from the US Census.
    A header row is not included. The columns are:
    
    * age
    * sex
    * educ
    * race
    * income
    * married
    '''
    path = Path(__file__).parent / 'california_pums.csv'
    
    # pragma: no cover to make local testing easier
    if not path.exists(): # pragma: no cover
        url = 'https://raw.githubusercontent.com/opendp/opendp/main/docs/source/data/PUMS_california_demographics_1000/data.csv'
        path.write_text(_http_get(url).decode())
    return path


def get_france_lfs_path() -> Path:
    '''
    Returns the path to a CSV derived from the
    `EU Labor Force Survey <https://ec.europa.eu/eurostat/web/microdata/public-microdata/labour-force-survey>`_. First row contains the column names.

    Code developed to work with a public microdata set like this could also be used with the scientific use files, and we believe that differential privacy would be a good tool to ensure that statistics derived from scientific use files could not inadvertantly reveal personal information.

    To reduce the download size for the tutorial, we've `preprocessed <https://github.com/opendp/dp-test-datasets/blob/main/data/eurostat/README.ipynb>`_ the data by selecting a single country (France), dropping unused columns, sampling a subset of the rows, and concatenating the result into a single CSV. The code we'll present in the tutorials could be run on the original public microdata, or for that matter, the full private scientific use files.
    '''
    from io import BytesIO
    from zipfile import ZipFile
    path = Path(__file__).parent / 'v2_france_lfs.csv'

    # pragma: no cover to make local testing easier
    if not path.exists(): # pragma: no cover
        url = 'https://raw.githubusercontent.com/opendp/dp-test-datasets/refs/heads/main/data/eurostat/V2_FR_LFS.csv.zip'
        france_lfs_bytes = _http_get(url)
        with ZipFile(BytesIO(france_lfs_bytes)) as data_zip:
            path.write_text(data_zip.open('V2_FR_LFS.csv').read().decode())
    return path


"""Releases all queries in a fixed workload."""

from dataclasses import dataclass
from typing import Any, TYPE_CHECKING

from opendp._internal import _new_pure_function
from opendp._lib import import_optional_dependency
from opendp.extras.mbi._utilities import (
    get_associated_metric,
    then_noise_marginals,
    weight_marginals,
    make_stable_marginals,
    Algorithm,
    Count,
    OnewayType,
    ONEWAY_UNKEYED
)
from opendp.mod import (
    FrameDistance,
    LazyFrameDomain,
    Measure,
    Measurement,
    binary_search_chain,
)

if TYPE_CHECKING:  # pragma: no cover
    from opendp.extras.polars import Bound


@dataclass(kw_only=True, frozen=True)
class Fixed(Algorithm):
    """Releases all queries in a fixed workload."""

    queries: list[Count]
    """Workload of queries."""
    oneway: OnewayType = ONEWAY_UNKEYED
    """Only fit one-way marginals for columns missing keys.
    
    The fixed algorithm differs from other algorithms
    in that it only estimates marginals with missing keys, 
    not all unknown first-order marginals.
    """

    def __post_init__(self):
        super().__post_init__()

        if not self.queries:
            raise ValueError("queries must have at least one element")

        if not all(isinstance(q, Count) for q in self.queries):
            raise ValueError("queries must be of type Count")

    def make_marginals(
        self,
        input_domain: LazyFrameDomain,
        input_metric: FrameDistance,
        output_measure: Measure,
        d_in: list["Bound"],
        d_out: float,
        *,
        marginals: dict[tuple[str, ...], Any],
        model: Any,  # MarkovRandomField
    ):
        """Implements a "Fixed" algorithm over ordinal data.

        The algorithm estimates fixed cliques.

        :param input_domain: domain of input data
        :param input_metric: how to compute distance between datasets
        :param output_measure: how to measure privacy of release
        :param d_in: distance between adjacent input datasets
        :param d_out: upper bound on the privacy loss
        :param marginals: prior marginal releases
        :param model: warm-start fit of MarkovRandomField
        """
        import_optional_dependency("mbi")
        from mbi import MarkovRandomField  # type: ignore[import-untyped,import-not-found]

        if not isinstance(model, MarkovRandomField):
            raise ValueError("model must be a MarkovRandomField")

        lp_metric = get_associated_metric(output_measure)
        cliques = [q.by for q in self.queries]
        weights = [q.weight for q in self.queries]

        def make(scale: float) -> Measurement:
            return make_stable_marginals(
                input_domain, input_metric, lp_metric, cliques  # type: ignore[arg-type]
            ) >> then_noise_marginals(
                output_measure, cliques, scale, weights
            )  # type: ignore[return-type]

        m_marginals = binary_search_chain(make, d_in, d_out, T=float)

        def function(
            new_releases: list,
        ) -> tuple[dict[tuple[str, ...], Any], MarkovRandomField]:
            all_marginals = weight_marginals(marginals, *new_releases)

            new_model = self.estimator(
                model.domain,
                list(all_marginals.values()),
                potentials=model.potentials.expand(list(all_marginals.keys())),
            )
            return all_marginals, new_model

        return m_marginals >> _new_pure_function(function)



# type: ignore
class CompositionMeasure(MaxDivergence):
    def composability(  # |\label{line:composability}|
        self, adaptivity: Adaptivity
    ) -> Composability:
        if matches(adaptivity, Adaptivity.FullyAdaptive):
            raise "fully-adaptive composition is not currently supported for max-divergence"
        return Composability.Concurrent

    def compose(self, d_mids: Vec[Self_Distance]) -> Self_Distance:
        d_out = 0.0
        for d_mid in d_mids:
            d_out = d_out.inf_add(d_mid)
        return d_out



'''
The ``mod`` module provides the classes which implement the
`OpenDP Programming Framework <../../api/user-guide/programming-framework/index.html>`_,
as well as utilities for enabling features and finding parameter values.

The classes here correspond to other top-level modules: For example,
instances of :py:class:`opendp.mod.Domain` are either inputs or outputs for functions in :py:mod:`opendp.domains`.
'''
from __future__ import annotations
import ctypes
from dataclasses import asdict
from typing import Any, Literal, Sequence, Type, TypeVar, Union, Callable, Optional, overload, TYPE_CHECKING, cast
import importlib
import json

from opendp._lib import AnyMeasurement, AnyTransformation, AnyDomain, AnyMetric, AnyMeasure, AnyFunction, AnyOdometer, import_optional_dependency, get_opendp_version


# https://mypy.readthedocs.io/en/stable/runtime_troubles.html#import-cycles
if TYPE_CHECKING:
    from opendp.typing import RuntimeType # pragma: no cover


__all__ = [
    'Measurement',
    'Transformation',
    'Odometer',
    'Queryable',
    'OdometerQueryable',
    'Function',
    'Domain',
    'AtomDomain',
    'OptionDomain',
    'VectorDomain',
    'SeriesDomain',
    'LazyFrameDomain',
    'ExtrinsicDomain',
    'Metric',
    'SymmetricIdDistance',
    'ChangeOneIdDistance',
    'FrameDistance',
    'Measure',
    'ApproximateDivergence',
    'PrivacyProfile',
    '_PartialConstructor',
    'UnknownTypeException',
    'OpenDPException',
    'GLOBAL_FEATURES',
    'enable_features',
    'disable_features',
    'assert_features',
    'binary_search_chain',
    'binary_search_param',
    'binary_search',
    'exponential_bounds_search',
    'serialize',
    'deserialize',
    '__version__',
]

class Measurement(ctypes.POINTER(AnyMeasurement)): # type: ignore[misc]
    """A differentially private unit of computation.
    A measurement contains a function and a privacy relation.
    The function releases a differentially-private release.
    The privacy relation maps from an input metric to an output measure.

    See the `Measurement <../../api/user-guide/programming-framework/core-structures.html#measurement>`_
    section in the Programming Framework docs for more context.

    Functions for creating measurements are in :py:mod:`opendp.measurements`.

    :example:

    >>> import opendp.prelude as dp
    >>> dp.enable_features("contrib")

    >>> # create an instance of Measurement using a constructor from the meas module
    >>> laplace = dp.m.make_laplace(
    ...     dp.atom_domain(T=int), dp.absolute_distance(T=int),
    ...     scale=2.)
    >>> laplace
    Measurement(
        input_domain   = AtomDomain(T=i32),
        input_metric   = AbsoluteDistance(i32),
        output_measure = MaxDivergence)

    >>> # invoke the measurement (invoke and __call__ are equivalent)
    >>> print('explicit: ', laplace.invoke(100))  # -> 101   # doctest: +ELLIPSIS
    explicit: ...
    >>> print('concise: ', laplace(100))  # -> 99            # doctest: +ELLIPSIS
    concise: ...
    >>> # check the measurement's relation at
    >>> #     (1, 0.5): (AbsoluteDistance<u32>, MaxDivergence)
    >>> assert laplace.check(1, 0.5)

    >>> # chain with a transformation from the trans module
    >>> chained = (
    ...     (dp.vector_domain(dp.atom_domain(T=int)), dp.symmetric_distance()) >>
    ...     dp.t.then_count() >>
    ...     laplace
    ... )

    >>> # the resulting measurement has the same features
    >>> print('dp count: ', chained([1, 2, 3]))  # -> 4     # doctest: +ELLIPSIS
    dp count: ...

    >>> # check the chained measurement's relation at
    >>> #     (1, 0.5): (SymmetricDistance, MaxDivergence)
    >>> assert chained.check(1, 0.5)
    """
    _type_ = AnyMeasurement

    def __call__(self, arg):
        from opendp.core import measurement_invoke
        return measurement_invoke(self, arg)

    def invoke(self, arg):
        """Create a differentially-private release with `arg`.

        If `self` is (d_in, d_out)-close, then each invocation of this function is a d_out-DP release. 
        
        :param arg: Input to the measurement.
        :return: differentially-private release
        :raises OpenDPException: packaged error from the core OpenDP library
        """
        from opendp.core import measurement_invoke
        return measurement_invoke(self, arg)

    def map(self, d_in):
        """Map an input distance `d_in` to an output distance.
        
        :param d_in: Input distance
        """
        from opendp.core import measurement_map
        return measurement_map(self, d_in)

    def check(self, d_in, d_out, *, debug=False) -> bool:
        """Check if the measurement is (`d_in`, `d_out`)-close.
        If true, implies that if the distance between inputs is at most `d_in`, then the privacy usage is at most `d_out`.
        See also :func:`~Transformation.check`, a similar check for transformations.
        
        :param d_in: Distance in terms of the input metric.
        :param d_out: Distance in terms of the output measure.
        :param debug: Enable to raise Exceptions to help identify why the privacy relation failed.
        :return: If True, a release is differentially private at `d_in`, `d_out`.
        :rtype: bool
        """
        from opendp.core import measurement_check

        if debug:
            return measurement_check(self, d_in, d_out)

        try:
            return measurement_check(self, d_in, d_out)
        except OpenDPException as err:
            if err.variant == "RelationDebug":
                return False # pragma: no cover
            raise

    def __rshift__(self, other: Union["Function", "Transformation", Callable]) -> "Measurement":
        if isinstance(other, Transformation):
            other = other.function

        if not isinstance(other, Function):
            if not callable(other):
                raise ValueError(f'Expected a callable instead of {other}')  # pragma: no cover
            from opendp.core import new_function
            other = new_function(other, TO="ExtrinsicObject")

        from opendp.combinators import make_chain_pm
        return make_chain_pm(other, self)

    @property
    def input_domain(self) -> "Domain":
        '''
        Input domain of measurement
        '''
        from opendp.core import measurement_input_domain
        return measurement_input_domain(self)
    
    @property
    def input_metric(self) -> "Metric":
        '''
        Input metric of measurement
        '''
        from opendp.core import measurement_input_metric
        return measurement_input_metric(self)

    @property
    def input_space(self) -> tuple["Domain", "Metric"]:
        '''
        Input space of measurement
        '''
        return self.input_domain, self.input_metric
    
    @property
    def output_measure(self) -> "Measure":
        '''
        Output measure of measurement
        '''
        from opendp.core import measurement_output_measure
        return measurement_output_measure(self)
    
    @property
    def function(self) -> "Function":
        '''
        Function of measurement
        '''
        from opendp.core import measurement_function
        return measurement_function(self)
    
    @property
    def input_distance_type(self) -> Union["RuntimeType", str]:
        """Retrieve the distance type of the input metric.
        This may be any integral type for dataset metrics, or any numeric type for sensitivity metrics.
        
        :return: distance type
        """
        from opendp.core import measurement_input_distance_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(measurement_input_distance_type(self))

    @property
    def output_distance_type(self) -> Union["RuntimeType", str]:
        """Retrieve the distance type of the output measure.
        This is the type that the budget is expressed in.
        
        :return: distance type
        """
        from opendp.core import measurement_output_distance_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(measurement_output_distance_type(self))

    @property
    def input_carrier_type(self) -> Union["RuntimeType", str]:
        """Retrieve the carrier type of the input domain.
        Any member of the input domain is a member of the carrier type.
        
        :return: carrier type
        """
        from opendp.core import measurement_input_carrier_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(measurement_input_carrier_type(self))

    def __del__(self):
        try:
            from opendp.core import _measurement_free
            _measurement_free(self)
        except (ImportError, TypeError): # pragma: no cover
            # an example error that this catches:
            #   ImportError: sys.meta_path is None, Python is likely shutting down
            pass
    
    def __repr__(self) -> str:
        return f"""Measurement(
    input_domain   = {self.input_domain},
    input_metric   = {self.input_metric},
    output_measure = {self.output_measure})"""

    def __iter__(self):
        # this overrides the implementation of __iter__ on POINTER, 
        # which yields infinitely on zero-sized types
        raise ValueError("Measurement does not support iteration")
    

class Odometer(ctypes.POINTER(AnyOdometer)): # type: ignore[misc]
    """A differentially private unit of computation with no privacy limit.
    An Odometer can be invoked with a dataset to return an OdometerQueryable.

    Differentially private queries (measurements) may be passed to the queryable,
    and ``.privacy_loss(d_in)`` can be called to check the current privacy usage.

    :example:

    >>> import opendp.prelude as dp
    >>> dp.enable_features("contrib")
    ...
    >>> # create a measurement that responds honestly with probability 0.6
    >>> meas_rr = dp.m.make_randomized_response_bool(prob=0.6)
    ...
    >>> # create an Odometer representing a fully adaptive compositor
    >>> odometer = dp.c.make_fully_adaptive_composition(
    ...     input_domain=meas_rr.input_domain,
    ...     input_metric=meas_rr.input_metric,
    ...     output_measure=meas_rr.output_measure,
    ... )
    ...
    >>> # invoke the odometer to get a queryable
    >>> data = True # a trivial boolean dataset
    >>> qbl_comp = odometer(data)
    ...
    >>> # evaluate the queryable (eval and __call__ are equivalent)
    >>> print("Release:", qbl_comp(meas_rr))  # doctest: +ELLIPSIS
    Release: ...
    >>> # odometers can be repeatedly invoked without any limit on the privacy loss
    >>> print("Release:", qbl_comp(meas_rr))  # doctest: +ELLIPSIS
    Release: ...
    >>> # The odometer's privacy consumption (in terms of ε) can be checked at any time.
    >>> # Input dataset may differ by discrete distance 1.
    >>> qbl_comp.privacy_loss(1)
    0.8109302162163288
    """
    _type_ = AnyOdometer

    def __call__(self, arg):
        from opendp.core import odometer_invoke
        return odometer_invoke(self, arg)

    def invoke(self, arg):
        """Create a differentially-private release with `arg`.

        If `self` is (d_in, d_out)-close, then each invocation of this function is a d_out-DP release. 
        
        :param arg: Input to the measurement.
        :return: differentially-private release
        :raises OpenDPException: packaged error from the core OpenDP library
        """
        from opendp.core import odometer_invoke
        return odometer_invoke(self, arg)

    @property
    def input_domain(self) -> "Domain":
        '''
        Input domain of odometer
        '''
        from opendp.core import odometer_input_domain
        return odometer_input_domain(self)
    
    @property
    def input_metric(self) -> "Metric":
        '''
        Input metric of odometer
        '''
        from opendp.core import odometer_input_metric
        return odometer_input_metric(self)
    
    @property
    def input_space(self) -> tuple["Domain", "Metric"]:
        '''
        Input domain and metric of odometer
        '''
        return self.input_domain, self.input_metric
    
    @property
    def output_measure(self) -> "Measure":
        '''
        Output measure of odometer
        '''
        from opendp.core import odometer_output_measure
        return odometer_output_measure(self)
    
    @property
    def input_distance_type(self) -> Union["RuntimeType", str]:
        """Retrieve the distance type of the input metric.
        This may be any integral type for dataset metrics, or any numeric type for sensitivity metrics.
        
        :return: distance type
        """
        return self.input_metric.distance_type

    @property
    def output_distance_type(self) -> Union["RuntimeType", str]:
        """Retrieve the distance type of the output measure.
        This is the type that the budget is expressed in.
        
        :return: distance type
        """
        return self.output_measure.distance_type

    @property
    def input_carrier_type(self) -> Union["RuntimeType", str]:
        """Retrieve the carrier type of the input domain.
        Any member of the input domain is a member of the carrier type.
        
        :return: carrier type
        """
        return self.input_domain.carrier_type

    def __del__(self):
        try:
            from opendp.core import _odometer_free
            _odometer_free(self)
        except (ImportError, TypeError): # pragma: no cover
            # ImportError: sys.meta_path is None, Python is likely shutting down
            # TypeError: similar setting as above
            pass
    
    def __repr__(self) -> str:
        return f"""Odometer(
    input_domain   = {self.input_domain},
    input_metric   = {self.input_metric},
    output_measure = {self.output_measure})"""

    def __iter__(self):
        # this overrides the implementation of __iter__ on POINTER, 
        # which yields infinitely on zero-sized types
        raise ValueError("Odometer does not support iteration")


class Transformation(ctypes.POINTER(AnyTransformation)): # type: ignore[misc]
    """A non-differentially private unit of computation.
    A transformation contains a function and a stability relation.
    The function maps from an input domain to an output domain.
    The stability relation maps from an input metric to an output metric.

    See the `Transformation <../../api/user-guide/programming-framework/core-structures.html#transformation>`_
    section in the Programming Framework docs for more context.

    Functions for creating transformations are in :py:mod:`opendp.transformations`.

    :example:

    >>> import opendp.prelude as dp
    >>> dp.enable_features("contrib")

    >>> # create an instance of Transformation using a constructor from the trans module
    >>> input_space = (dp.vector_domain(dp.atom_domain(T=int)), dp.symmetric_distance())
    >>> count = input_space >> dp.t.then_count()
    >>> count
    Transformation(
        input_domain   = VectorDomain(AtomDomain(T=i32)),
        output_domain  = AtomDomain(T=i32),
        input_metric   = SymmetricDistance(),
        output_metric  = AbsoluteDistance(i32))

    >>> count.input_space
    (VectorDomain(AtomDomain(T=i32)), SymmetricDistance())
    
    >>> # invoke the transformation (invoke and __call__ are equivalent)
    >>> count.invoke([1, 2, 3])
    3
    >>> count([1, 2, 3])
    3
    >>> # check the transformation's relation at
    >>> #     (1, 1): (SymmetricDistance, AbsoluteDistance<u32>)
    >>> assert count.check(1, 1)

    >>> # chain with more transformations from the trans module
    >>> chained = (
    ...     dp.t.make_split_lines() >>
    ...     dp.t.then_cast_default(TOA=int) >>
    ...     count
    ... )

    >>> # the resulting transformation has the same features
    >>> chained("1\\n2\\n3")
    3
    >>> assert chained.check(1, 1)  # both chained transformations were 1-stable
    """
    _type_ = AnyTransformation

    def invoke(self, arg):
        """Execute a non-differentially-private query with `arg`.

        :param arg: Input to the transformation.
        :return: non-differentially-private answer
        :raises OpenDPException: packaged error from the core OpenDP library
        """
        from opendp.core import transformation_invoke
        return transformation_invoke(self, arg) 

    def __call__(self, arg):
        from opendp.core import transformation_invoke
        return transformation_invoke(self, arg)

    def map(self, d_in):
        """Map an input distance `d_in` to an output distance.
        
        :param d_in: Input distance. An upper bound on how far apart neighboring datasets can be with respect to the input metric
        """
        from opendp.core import transformation_map
        return transformation_map(self, d_in)

    def check(self, d_in, d_out, *, debug=False) -> bool:
        """Check if the transformation is (`d_in`, `d_out`)-close.
        If true, implies that if the distance between inputs is at most `d_in`, then the distance between outputs is at most `d_out`.
        See also :func:`~Measurement.check`, a similar check for measurements.

        :param d_in: Distance in terms of the input metric.
        :param d_out: Distance in terms of the output metric.
        :param debug: Enable to raise Exceptions to help identify why the stability relation failed.
        :return: True if the relation passes. False if the relation failed.
        :rtype: bool
        :raises OpenDPException: packaged error from the core OpenDP library
        """
        from opendp.core import transformation_check

        if debug:
            return transformation_check(self, d_in, d_out)

        try:
            return transformation_check(self, d_in, d_out)
        except OpenDPException as err: # pragma: no cover
            if err.variant == "RelationDebug":
                return False
            raise

    @overload
    def __rshift__(self, other: "Transformation") -> "Transformation":
        ...

    @overload
    def __rshift__(self, other: "Measurement") -> "Measurement":
        ...

    @overload
    def __rshift__(self, other: "Odometer") -> "Odometer":
        ...

    @overload
    def __rshift__(self, other: "_PartialConstructor") -> Union["Transformation", "Measurement", "Odometer"]:
        ...

    def __rshift__(self, other: Union["Measurement", "Transformation", "_PartialConstructor"]) -> Union["Measurement", "Transformation", "_PartialConstructor", "PartialChain"]:  # type: ignore[name-defined] # noqa F821
        if isinstance(other, Measurement):
            from opendp.combinators import make_chain_mt
            return make_chain_mt(other, self)

        if isinstance(other, Transformation):
            from opendp.combinators import make_chain_tt
            return make_chain_tt(other, self)
        
        if isinstance(other, _PartialConstructor):
            return self >> other(self.output_domain, self.output_metric) # type: ignore[call-arg]

        from opendp.context import PartialChain
        if isinstance(other, PartialChain):
            return PartialChain(lambda x: self >> other.partial(x))

        raise ValueError(f"rshift expected a measurement or transformation, got {other}")  # pragma: no cover


    @property
    def input_domain(self) -> "Domain":
        '''
        Input domain of transformation
        '''
        from opendp.core import transformation_input_domain
        return transformation_input_domain(self)
    

    @property
    def output_domain(self) -> "Domain":
        '''
        Output domain of transformation
        '''
        from opendp.core import transformation_output_domain
        return transformation_output_domain(self)
    

    @property
    def input_metric(self) -> "Metric":
        '''
        Input metric of transformation
        '''
        from opendp.core import transformation_input_metric
        return transformation_input_metric(self)
    
    @property
    def output_metric(self) -> "Metric":
        '''
        Ouput metric of transformation
        '''
        from opendp.core import transformation_output_metric
        return transformation_output_metric(self)
    
    @property
    def input_space(self) -> tuple["Domain", "Metric"]:
        '''
        Input space of transformation
        '''
        return self.input_domain, self.input_metric
    
    @property
    def output_space(self) -> tuple["Domain", "Metric"]:
        '''
        Output space of transformation
        '''
        return self.output_domain, self.output_metric
    
    @property
    def function(self) -> "Function":
        '''
        Function of transformation
        '''
        from opendp.core import transformation_function
        return transformation_function(self)

    @property
    def input_distance_type(self) -> Union["RuntimeType", str]:
        """Retrieve the distance type of the input metric.
        This may be any integral type for dataset metrics, or any numeric type for sensitivity metrics.

        :return: distance type
        """
        from opendp.core import transformation_input_distance_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(transformation_input_distance_type(self))

    @property
    def output_distance_type(self) -> Union["RuntimeType", str]:
        """Retrieve the distance type of the output metric.
        This may be any integral type for dataset metrics, or any numeric type for sensitivity metrics.

        :return: distance type
        """
        from opendp.core import transformation_output_distance_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(transformation_output_distance_type(self))
    
    @property
    def input_carrier_type(self) -> Union["RuntimeType", str]:
        """Retrieve the carrier type of the input domain.
        Any member of the input domain is a member of the carrier type.

        :return: carrier type
        """
        from opendp.core import transformation_input_carrier_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(transformation_input_carrier_type(self))

    def __del__(self):
        try:
            from opendp.core import _transformation_free
            _transformation_free(self)
        except (ImportError, TypeError): # pragma: no cover
            # an example error that this catches:
            #   ImportError: sys.meta_path is None, Python is likely shutting down
            pass

    def __repr__(self) -> str:
        return f"""Transformation(
    input_domain   = {self.input_domain},
    output_domain  = {self.output_domain},
    input_metric   = {self.input_metric},
    output_metric  = {self.output_metric})"""
    
    def __iter__(self):
        raise ValueError("Transformation does not support iteration")


Transformation = cast(Type[Transformation], Transformation) # type: ignore[misc]

class Queryable:
    '''
    Queryables are used for interactive mechanisms like :ref:`adaptive composition <adaptive-composition>`.

    Queryables can be created with :py:func:`make_adaptive_composition <opendp.combinators.make_adaptive_composition>`
    or :py:func:`new_queryable <opendp.core.new_queryable>`.
    '''
    def __init__(self, value, query_type):
        self.value = value
        self.query_type = query_type

    def __call__(self, query):
        from opendp.core import queryable_eval
        return queryable_eval(self.value, query)
    
    def __repr__(self) -> str:
        return f"Queryable(Q={self.query_type})"

class OdometerQueryable:
    '''
    Odometer Queryables are used for instances of odometers like :ref:`fully adaptive composition <fully-adaptive-composition>`.

    Can be created via :py:func:`make_fully_adaptive_composition <opendp.combinators.make_fully_adaptive_composition>`.
    '''
    def __init__(self, value):
        self.value = value

    def __call__(self, query):
        from opendp.core import odometer_queryable_invoke
        return odometer_queryable_invoke(self.value, query)
    
    def invoke(self, query):
        from opendp.core import odometer_queryable_invoke
        return odometer_queryable_invoke(self.value, query)
    
    def privacy_loss(self, d_in):
        from opendp.core import odometer_queryable_privacy_loss
        return odometer_queryable_privacy_loss(self.value, d_in)

    def __repr__(self) -> str:
        from opendp.core import odometer_queryable_invoke_type, odometer_queryable_privacy_loss_type
        from opendp.typing import RuntimeType
        Q = RuntimeType.parse(odometer_queryable_invoke_type(self.value))
        QB = RuntimeType.parse(odometer_queryable_privacy_loss_type(self.value))
        return f"OdometerQueryable(Q={Q}, QB={QB})"


class Function(ctypes.POINTER(AnyFunction)): # type: ignore[misc]
    '''
    See the `Function <../../api/user-guide/programming-framework/supporting-elements.html#function>`_
    section in the Programming Framework docs for more context.
    '''
    _type_ = AnyFunction

    def __call__(self, arg):
        from opendp.core import function_eval
        return function_eval(self, arg)
    
    def __del__(self):
        try:
            from opendp.core import _function_free
            _function_free(self)
        except (ImportError, TypeError): # pragma: no cover
            # an example error that this catches:
            #   ImportError: sys.meta_path is None, Python is likely shutting down
            pass

    def __iter__(self):
        raise ValueError("Function does not support iteration")


D = TypeVar("D")

class Domain(ctypes.POINTER(AnyDomain)): # type: ignore[misc]
    '''
    See the `Domain <../../api/user-guide/programming-framework/supporting-elements.html#domain>`_
    section in the Programming Framework docs for more context.

    Functions for creating domains are in :py:mod:`opendp.domains`.
    '''
    
    # for documentation see https://docs.python.org/3/library/ctypes.html#ctypes._Pointer._type_
    _type_ = AnyDomain

    def member(self, val) -> bool:
        '''
        Check if ``val`` is a member of the domain.
        
        :param val: a value to be checked for membership in `self`
        '''
        try:
            from opendp.domains import _member
            return _member(self, val)
        except Exception as e:
            from warnings import warn
            warn(f'Value ({val}) does not belong to carrier type of {self}. Details: {e}')
            return False


    @property
    def type(self) -> Union["RuntimeType", str]:
        '''
        Type of domain
        '''
        from opendp.domains import domain_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(domain_type(self))
    
    @property
    def carrier_type(self) -> Union["RuntimeType", str]:
        '''
        Carrier type of domain
        '''
        from opendp.domains import domain_carrier_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(domain_carrier_type(self))
    
    def __repr__(self) -> str:
        from opendp.domains import domain_debug
        return domain_debug(self)
    
    def __del__(self):
        try:
            from opendp.domains import _domain_free
            _domain_free(self)
        except (ImportError, TypeError): # pragma: no cover
            # an example error that this catches:
            #   ImportError: sys.meta_path is None, Python is likely shutting down
            pass
    
    def __eq__(self, other) -> bool:
        from opendp.domains import _domain_equal

        if not isinstance(other, Domain):
            return False
        
        return _domain_equal(self, other)
    
    def __hash__(self) -> int:
        return hash(str(self))
    
    def __iter__(self):
        raise ValueError("Domain does not support iteration")
    
    def cast(self, type_: Type[D]) -> D:
        """Retrieve the descriptor as the prescribed type, or error."""
        if not (
            isinstance(self, ExtrinsicDomain)
            and isinstance(descriptor := self.descriptor, type_)
        ):
            raise ValueError(f"domain descriptor must be a {type_.__name__}")
        return descriptor


class AtomDomain(Domain):
    '''The domain of all values of a given atomic type.

    Create an instance of this domain with :py:func:`opendp.domains.atom_domain`.

    If bounds are set, then the domain is restricted to the bounds.
    If nullable is set, then null value(s) are included in the domain.
    '''

    _type_ = AnyDomain
    
    @property
    def bounds(self) -> tuple[float, float]:
        '''Bounds of the domain, if they exist'''
        from opendp.domains import _atom_domain_get_bounds_closed
        return _atom_domain_get_bounds_closed(self)
    
    @property
    def nan(self) -> bool:
        """Whether the domain includes NaN values
        
        Only relevant when the carrier type is a floating point type.
        All other types will always return ``False``.
        """
        from opendp.domains import _atom_domain_nan
        return _atom_domain_nan(self)
    

class OptionDomain(Domain):
    '''A domain whose members are either members of the ``element_domain``, or ``None``.

    Create an instance of this domain with :py:func:`opendp.domains.option_domain`.

    The element domain is the domain of non-null values.
    '''

    _type_ = AnyDomain

    @property
    def element_domain(self) -> Domain:
        '''Domain of non-null values'''
        from opendp.domains import _option_domain_get_element_domain
        return _option_domain_get_element_domain(self)


class VectorDomain(Domain):
    '''``VectorDomain`` describes the domain of all vectors whose elements are members of a given domain.
    
    Create an instance of this domain with :py:func:`opendp.domains.vector_domain`.
    '''
    _type_ = AnyDomain
    
    @property
    def element_domain(self) -> Domain:
        '''Domain of elements in the vector'''
        from opendp.domains import _vector_domain_get_element_domain
        return _vector_domain_get_element_domain(self)
    
    @property
    def size(self) -> Optional[int]:
        '''Size of vectors in the domain, if it is fixed'''
        from opendp.domains import _vector_domain_get_size
        return _vector_domain_get_size(self)


class SeriesDomain(Domain):
    '''``SeriesDomain`` describes the domain of all polars Series.
    
    Create an instance of this domain with :py:func:`opendp.domains.series_domain`.
    '''
    _type_ = AnyDomain

    @property
    def name(self) -> str:
        '''Name of series in the domain'''
        from opendp.domains import _series_domain_get_name
        return _series_domain_get_name(self)
    
    @property
    def element_domain(self) -> Domain:
        '''Domain of non-null elements in the series'''
        from opendp.domains import _series_domain_get_element_domain
        return _series_domain_get_element_domain(self)
    
    @property
    def nullable(self) -> bool:
        '''Whether series in the domain may include null values'''
        from opendp.domains import _series_domain_get_nullable
        return _series_domain_get_nullable(self)
    
class LazyFrameDomain(Domain):
    '''``LazyFrameDomain`` describes the domain of all polars LazyFrames.
    
    Create an instance of this domain with :py:func:`opendp.domains.lazyframe_domain`.
    '''
    _type_ = AnyDomain

    @property
    def columns(self) -> list[str]:
        '''List of column names in the frame'''
        from opendp.domains import _lazyframe_domain_get_columns
        return _lazyframe_domain_get_columns(self)

    def get_series_domain(self, name: str) -> SeriesDomain:
        '''Retrieve the series domain with the given name'''
        from opendp.domains import _lazyframe_domain_get_series_domain
        return _lazyframe_domain_get_series_domain(self, name)
    
    def get_margin(self, by: Sequence[Any]):
        '''Get the margin descriptor of the frame when grouped by the given columns'''
        from opendp.domains import _lazyframe_domain_get_margin
        from opendp._convert import _check_polars_by
        _check_polars_by(by)

        return _lazyframe_domain_get_margin(self, by)


class ExtrinsicDomain(Domain):
    '''A user-defined domain.'''

    _type_ = AnyDomain
        
    @property
    def descriptor(self) -> Any:
        '''
        Descriptor of domain. Used to retrieve the descriptor associated with domains defined in Python 
        '''
        from opendp.domains import _extrinsic_domain_descriptor
        return _extrinsic_domain_descriptor(self)


class Metric(ctypes.POINTER(AnyMetric)): # type: ignore[misc]
    '''
    See the `Metric <../../api/user-guide/programming-framework/supporting-elements.html#metric>`_
    section in the Programming Framework docs for more context.

    Functions for creating metrics are in :py:mod:`opendp.metrics`.
    '''
    _type_ = AnyMetric

    @property
    def type(self):
        '''
        Type of metric
        '''
        from opendp.metrics import metric_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(metric_type(self))
    
    @property
    def distance_type(self) -> Union["RuntimeType", str]:
        '''
        Distance type of metric
        '''
        from opendp.metrics import metric_distance_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(metric_distance_type(self))

    def __repr__(self) -> str:
        from opendp.metrics import metric_debug
        return metric_debug(self)
    
    def __del__(self):
        try:
            from opendp.metrics import _metric_free
            _metric_free(self)
        except (ImportError, TypeError): # pragma: no cover
            # an example error that this catches:
            #   ImportError: sys.meta_path is None, Python is likely shutting down
            pass
    
    def __eq__(self, other) -> bool:
        from opendp.metrics import _metric_equal

        if not isinstance(other, Metric):
            return False
        
        return _metric_equal(self, other)
    
    def __hash__(self) -> int:
        return hash(str(self))
    
    def __iter__(self):
        raise ValueError("Metric does not support iteration")
    
    def cast(self, type_: Type[D]) -> D:
        """Retrieve the descriptor as the prescribed type, or error."""
        if not (
            isinstance(self, ExtrinsicDistance)
            and isinstance(descriptor := self.descriptor, type_)
        ):
            raise ValueError(f"metric descriptor must be a {type_.__name__}, found {self}")
        return descriptor


class ExtrinsicDistance(Metric):
    '''A user-defined metric.'''

    _type_ = AnyMetric
        
    @property
    def descriptor(self) -> Any:
        '''
        Descriptor of domain. Used to retrieve the descriptor associated with domains defined in Python 
        '''
        from opendp.metrics import _extrinsic_metric_descriptor
        return _extrinsic_metric_descriptor(self)


class FrameDistance(Metric):
    '''``FrameDistance`` is a higher-order metric that contains multiple distance bounds for different groupings of data.'''

    _type_ = AnyMetric
    
    @property
    def inner_metric(self) -> Metric:
        '''Bounds of the domain, if they exist'''
        from opendp.metrics import _frame_distance_get_inner_metric
        return _frame_distance_get_inner_metric(self)
    
class SymmetricIdDistance(Metric):
    '''``SymmetricIdDistance`` is a metric for measuring the distance between the identifiers of two datasets.
    
    The metric counts the number of identifiers that must be added or removed to make the two datasets equal.
    '''

    _type_ = AnyMetric
    
    @property
    def identifier(self):
        '''The name of the column storing identifiers'''
        from opendp.metrics import _symmetric_id_distance_get_identifier
        return _symmetric_id_distance_get_identifier(self)
    

class ChangeOneIdDistance(Metric):
    '''``ChangeOneIdDistance`` is a metric for measuring the distance between the identifiers of two datasets.
    
    The metric counts the number of identifiers that must be changed to make the two datasets equal.
    '''

    _type_ = AnyMetric
    
    @property
    def identifier(self):
        '''The name of the column storing identifiers'''
        from opendp.metrics import _change_one_id_distance_get_identifier
        return _change_one_id_distance_get_identifier(self)
    
class Measure(ctypes.POINTER(AnyMeasure)): # type: ignore[misc]
    '''
    See the `Measure <../../api/user-guide/programming-framework/supporting-elements.html#measure>`_
    section in the Programming Framework docs for more context.

    Measures should be created with the functions in :py:mod:`opendp.measures`
    or :py:mod:`opendp.context`, for a higher-level interface:

    >>> import opendp.prelude as dp
    >>> measure, distance = dp.loss_of(epsilon=1.0)
    >>> measure, distance
    (MaxDivergence, 1.0)

    '''
    _type_ = AnyMeasure

    @property
    def type(self):
        '''
        Type of measure
        '''
        from opendp.measures import measure_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(measure_type(self))
    
    @property
    def distance_type(self) -> Union["RuntimeType", str]:
        '''
        Distance type of measure
        '''
        from opendp.measures import measure_distance_type
        from opendp.typing import RuntimeType
        return RuntimeType.parse(measure_distance_type(self))

    def __repr__(self):
        from opendp.measures import measure_debug
        return measure_debug(self)
    
    def __del__(self):
        try:
            from opendp.measures import _measure_free
            _measure_free(self)
        except (ImportError, TypeError): # pragma: no cover
            # an example error that this catches:
            #   ImportError: sys.meta_path is None, Python is likely shutting down
            pass

    def __eq__(self, other):
        from opendp.measures import _measure_equal

        if not isinstance(other, Measure):
            return False
        
        return _measure_equal(self, other)
    
    def __hash__(self) -> int:
        return hash(str(self))
    
    def __iter__(self):
        raise ValueError("Measure does not support iteration")



class ApproximateDivergence(Measure):
    '''``ApproximateDivergence`` is a privacy measure representing the divergence between two distributions 
    with respect to some inner privacy measure, except for some subset of possible outputs with probability mass no greater than delta.
    '''

    _type_ = AnyMeasure
    
    @property
    def inner_measure(self) -> Measure:
        from opendp.measures import _approximate_divergence_get_inner_measure
        return _approximate_divergence_get_inner_measure(self)
    

class PrivacyProfile(object):
    '''
    Given a profile function provided by the user,
    gives the epsilon corresponding to a given delta, and vice versa.

    :py:func:`new_privacy_profile <opendp.measures.new_privacy_profile>`
    should be used to create new instances.
    '''
    def __init__(self, curve):
        self.curve = curve

    def delta(self, epsilon):
        '''
        Returns the delta that corresponds to this epsilon.
        
        :param epsilon: Allowance for a multiplicative difference, or max divergence, in the distributions of releases on adjacent datasets
        '''
        from opendp._data import privacy_profile_delta
        return privacy_profile_delta(self.curve, epsilon)

    def epsilon(self, delta):
        '''
        Returns the epsilon that corresponds to this delta.
        
        :param delta: Allowance for an additive difference between the distributions of releases on adjacent datasets
        '''
        from opendp._data import privacy_profile_epsilon
        return privacy_profile_epsilon(self.curve, delta)
    

class _PartialConstructor(object):
    '''

    '''
    def __init__(self, constructor):
        self.constructor = constructor
        self.__opendp_dict__ = {}  # Not needed at runtime, but the definition prevents mypy errors.
    
    def __call__(self, input_domain: Domain, input_metric: Metric):
        return self.constructor(input_domain, input_metric)
    
    def __rshift__(self, other):
        return _PartialConstructor(lambda input_domain, input_metric: self(input_domain, input_metric) >> other) # pragma: no cover

    def __rrshift__(self, other):
        if isinstance(other, tuple):  
            domain, metric = other
            if isinstance(domain, Domain) and isinstance(metric, Metric):  
                return self(domain, metric)  
            
        raise TypeError(f"Cannot chain {type(self)} with {type(other)}")  # pragma: no cover


class UnknownTypeException(Exception):
    pass


class OpenDPException(Exception):
    """General exception for errors originating from the underlying OpenDP library.
    The variant attribute corresponds to `one of the following variants <https://github.com/opendp/opendp/blob/53ec58d01762ca5ceee08590d7e7b725bbdafcf6/rust/opendp/src/error.rs#L46-L87>`_ and can be matched on.
    Error variants may change in library updates.

    See `Rust ErrorVariant <https://docs.rs/opendp/latest/opendp/error/enum.ErrorVariant.html>`_ for values variant may take on.

    Run ``dp.enable_features('rust-stack-trace')`` to see wrapped Rust stack traces.
    """
    raw_traceback: Optional[str]

    def __init__(self, variant: str, message: Optional[str] = None, raw_traceback: Optional[str] = None):
        self.variant = variant
        self.message = message
        self.raw_traceback = raw_traceback

    def _raw_frames(self):
        import re
        return re.split(r"\s*[0-9]+: ", self.raw_traceback or "")
    
    def _frames(self):
        def _format_frame(frame):
            return "\n  ".join(line.strip() for line in frame.split("\n"))
        return [_format_frame(f) for f in self._raw_frames() if "opendp" in f]

    def _continued_stack_trace(self):
        # join and split by newlines because frames may be multi-line
        lines = "\n".join(self._frames()[::-1]).split('\n')
        return "Continued Rust stack trace:\n" + '\n'.join('    ' + line for line in lines)

    def __str__(self) -> str:
        '''
        >>> raw_traceback = """
        ... 0: top
        ... 1: opendp single line
        ... 2: opendp multi
        ...             line
        ... 3: bottom
        ... """
        >>> e = OpenDPException(variant='SomeVariant', message='my message', raw_traceback=raw_traceback)
        >>> dp.enable_features('rust-stack-trace')
        >>> print(e)
        Continued Rust stack trace:
            opendp multi
              line
            opendp single line
          SomeVariant("my message")
        >>> dp.disable_features('rust-stack-trace')
        >>> print(e)
        <BLANKLINE>
          SomeVariant("my message")
        '''
        response = ''
        if self.raw_traceback and 'rust-stack-trace' in GLOBAL_FEATURES:
            response += self._continued_stack_trace()
        response += '\n  ' + self.variant

        if self.message:
            response += f'("{self.message}")'
            
        return response


GLOBAL_FEATURES: set[str] = set()


def enable_features(*features: str) -> None:
    '''
    Allow the use of optional features. See :ref:`feature-listing` for details.
    '''
    GLOBAL_FEATURES.update(set(features))


def disable_features(*features: str) -> None:
    '''
    Disallow the use of optional features. See :ref:`feature-listing` for details.
    '''
    GLOBAL_FEATURES.difference_update(set(features))


def assert_features(*features: str) -> None:
    '''
    Check whether a given feature is enabled. See :ref:`feature-listing` for details.
    '''
    missing_features = [f for f in features if f not in GLOBAL_FEATURES]
    if missing_features:
        features_string = ', '.join(f'"{f}"' for f in features)
        raise OpenDPException(f"Attempted to use function that requires {features_string}, but not enabled. See https://github.com/opendp/opendp/discussions/304, then call enable_features({features_string})")


M = TypeVar("M", Transformation, Measurement)

def binary_search_chain(
        make_chain: Callable[[float], M],
        d_in: Any, d_out: Any,
        bounds: tuple[float, float] | None = None,
        T=None) -> M:
    """Find the highest-utility (`d_in`, `d_out`)-close Transformation or Measurement.
    
    Searches for the numeric parameter to `make_chain` that results in a computation
    that most tightly satisfies `d_out` when datasets differ by at most `d_in`,
    then returns the Transformation or Measurement corresponding to said parameter.

    See `binary_search_param` to retrieve the discovered parameter instead of the complete computation chain.

    :param make_chain: a function that takes a number and returns a Transformation or Measurement
    :param d_in: how far apart input datasets can be
    :param d_out: how far apart output datasets or distributions can be
    :param bounds: a 2-tuple of the lower and upper bounds on the input of `make_chain`
    :param T: type of argument to `make_chain`, one of {float, int}
    :return: a chain parameterized at the nearest passing value to the decision point of the relation
    :rtype: Union[Transformation, Measurement]
    :raises TypeError: if the type is not inferrable (pass T) or the type is invalid
    :raises ValueError: if the predicate function is constant, bounds cannot be inferred, or decision boundary is not within `bounds`.


    :example:

    Find a laplace measurement with the smallest noise scale that is still (d_in, d_out)-close.

    >>> import opendp.prelude as dp
    >>> dp.enable_features("floating-point", "contrib")
    ...
    >>> # The majority of the chain only needs to be defined once.
    >>> pre = (
    ...     dp.space_of(list[float]) >>
    ...     dp.t.then_impute_constant(0.0) >>
    ...     dp.t.then_clamp(bounds=(0., 1.)) >>
    ...     dp.t.then_resize(size=10, constant=0.) >>
    ...     dp.t.then_mean()
    ... )
    ...
    >>> # Find a value in `bounds` that produces a (`d_in`, `d_out`)-chain nearest the decision boundary.
    >>> # The lambda function returns the complete computation chain when given a single numeric parameter.
    >>> chain = dp.binary_search_chain(
    ...     lambda s: pre >> dp.m.then_laplace(scale=s), 
    ...     d_in=1, d_out=1.)
    ...
    >>> # The resulting computation chain is always (`d_in`, `d_out`)-close, but we can still double-check:
    >>> assert chain.check(1, 1.)


    Build a (2 neighboring, 1. epsilon)-close sized bounded sum with discrete_laplace(100.) noise.
    It should have the widest possible admissible clamping bounds (-b, b).

    >>> def make_sum(b):
    ...     space = dp.vector_domain(dp.atom_domain((-b, b)), 10_000), dp.symmetric_distance()
    ...     return space >> dp.t.then_sum() >> dp.m.then_laplace(100.)
    ...
    >>> # `meas` is a Measurement with the widest possible clamping bounds.
    >>> meas = dp.binary_search_chain(make_sum, d_in=2, d_out=1., bounds=(0, 10_000))
    ...
    >>> # If you want the discovered clamping bound, use `binary_search_param` instead.
    """
    return make_chain(binary_search_param(make_chain, d_in, d_out, bounds, T))


def binary_search_param(
        make_chain: Callable[[float], Union[Transformation, Measurement]],
        d_in: Any, d_out: Any,
        bounds: tuple[float, float] | None = None,
        T=None) -> float:
    """Solve for the ideal constructor argument to `make_chain`.
    
    Optimizes a parameterized chain `make_chain` within float or integer `bounds`,
    subject to the chained relation being (`d_in`, `d_out`)-close.

    :param make_chain: a function that takes a number and returns a Transformation or Measurement
    :param d_in: how far apart input datasets can be
    :param d_out: how far apart output datasets or distributions can be
    :param bounds: a 2-tuple of the lower and upper bounds on the input of `make_chain`
    :param T: type of argument to `make_chain`, one of {float, int}
    :return: the nearest passing value to the decision point of the relation
    :raises TypeError: if the type is not inferrable (pass T) or the type is invalid
    :raises ValueError: if the predicate function is constant, bounds cannot be inferred, or decision boundary is not within `bounds`.

    :example:

    >>> import opendp.prelude as dp
    ...
    >>> # Find a value in `bounds` that produces a (`d_in`, `d_out`)-chain nearest the decision boundary.
    >>> # The first argument is any function that returns your complete computation chain
    >>> #     when passed a single numeric parameter.
    ...
    >>> def make_fixed_laplace(scale):
    ...     # fixes the input domain and metric, but parameterizes the noise scale
    ...     return dp.m.make_laplace(dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float), scale)
    ...
    >>> scale = dp.binary_search_param(make_fixed_laplace, d_in=0.1, d_out=1.)
    >>> assert scale == 0.1
    >>> # Constructing the same chain with the discovered parameter will always be (0.1, 1.)-close.
    >>> assert make_fixed_laplace(scale).check(0.1, 1.)

    A policy research organization wants to know the smallest sample size necessary to release an "accurate" epsilon=1 DP mean income. 
    Determine the smallest dataset size such that, with 95% confidence, 
    the DP release differs from the clipped dataset's mean by no more than 1000. 
    Assume that neighboring datasets have a symmetric distance at most 2. 
    Also assume a clipping bound of 500,000.

    >>> # we first work out the necessary noise scale to satisfy the above constraints.
    >>> necessary_scale = dp.accuracy_to_laplacian_scale(accuracy=1000., alpha=.05)
    ...
    >>> # we then write a function that make a computation chain with a given data size
    >>> def make_mean(data_size):
    ...    return (
    ...        (dp.vector_domain(dp.atom_domain(bounds=(0., 500_000.)), data_size), dp.symmetric_distance()) >>
    ...        dp.t.then_mean() >> 
    ...        dp.m.then_laplace(necessary_scale)
    ...    )
    ...
    >>> # solve for the smallest dataset size that admits a (2 neighboring, 1. epsilon)-close measurement
    >>> dp.binary_search_param(
    ...     make_mean, 
    ...     d_in=2, d_out=1.,
    ...     bounds=(1, 1000000))
    1498
    """

    # one might think running scipy.optimize.brent* would be better, but 
    # 1. benchmarking showed no difference or minor regressions
    # 2. brentq is more complicated

    return binary_search(lambda param: make_chain(param).check(d_in, d_out), bounds, T)

# when return sign is false, only return float
@overload
def binary_search(
        predicate: Callable[[float], bool],
        bounds: tuple[float, float] | None = ...,
        T: Type[float] | None = ...,
        return_sign: Literal[False] = False) -> float:
    ...


# when setting return sign to true as a keyword argument, return both
@overload
def binary_search(
        predicate: Callable[[float], bool],
        bounds: tuple[float, float] | None = ...,
        T: Type[float] | None = ...,
        *, # see https://stackoverflow.com/questions/66435480/overload-following-optional-argument
        return_sign: Literal[True]) -> tuple[float, int]:
    ...

# when setting return sign to true as a positional argument, return both
@overload
def binary_search(
        predicate: Callable[[float], bool],
        bounds: tuple[float, float] | None,
        T: Type[float] | None,
        return_sign: Literal[True]) -> tuple[float, int]:
    ...


def binary_search(
        predicate: Callable[[float], bool],
        bounds: tuple[float, float] | None = None,
        T: Type[float] | None = None,
        return_sign: bool = False) -> float | tuple[float, int]:
    """Find the closest passing value to the decision boundary of `predicate`.

    If bounds are not passed, conducts an exponential search.
    
    :param predicate: a monotonic unary function from a number to a boolean
    :param bounds: a 2-tuple of the lower and upper bounds to the input of `predicate`
    :param T: type of argument to `predicate`, one of {float, int}
    :param return_sign: if True, also return the direction away from the decision boundary
    :return: the discovered parameter within the bounds
    :raises TypeError: if the type is not inferrable (pass T) or the type is invalid
    :raises ValueError: if the predicate function is constant, bounds cannot be inferred, or decision boundary is not within `bounds`.

    :example:

    >>> import opendp.prelude as dp
    >>> dp.binary_search(lambda x: x >= 5.)
    5.0
    >>> dp.binary_search(lambda x: x <= 5.)
    5.0
    >>> dp.binary_search(lambda x: x > 5, T=int)
    6
    >>> dp.binary_search(lambda x: x < 5, T=int)
    4

    Find epsilon usage of the gaussian(scale=1.) mechanism applied on a dp mean.
    Assume neighboring datasets differ by up to three additions/removals, and fix delta to 1e-8.

    >>> # build a histogram that emits float counts
    >>> input_space = dp.vector_domain(dp.atom_domain(bounds=(0., 100.)), 1000), dp.symmetric_distance()
    >>> dp_mean = dp.c.make_fix_delta(dp.c.make_zCDP_to_approxDP(
    ...     input_space >> dp.t.then_mean() >> dp.m.then_gaussian(1.)), 
    ...     1e-8
    ... )
    ...
    >>> dp.binary_search(
    ...     lambda d_out: dp_mean.check(3, (d_out, 1e-8)), 
    ...     bounds = (0., 1.))
    0.5235561269546629

    Find the L2 distance sensitivity of a histogram when neighboring datasets differ by up to 3 additions/removals.

    >>> histogram = dp.t.make_count_by_categories(
    ...     dp.vector_domain(dp.atom_domain(T=str)), dp.symmetric_distance(),
    ...     categories=["a"], MO=dp.L2Distance[int])
    ...
    >>> dp.binary_search(
    ...     lambda d_out: histogram.check(3, d_out), 
    ...     bounds = (0, 100))
    3
    """
    if bounds is None:
        bounds = exponential_bounds_search(predicate, T) # type: ignore

    if bounds is None:
        raise ValueError("unable to infer bounds")  # pragma: no cover

    if len(set(map(type, bounds))) != 1:
        raise TypeError("bounds must share the same type")  # pragma: no cover
    lower, upper = sorted(bounds)

    maximize = predicate(lower)  # if the lower bound passes, we should maximize
    minimize = predicate(upper)  # if the upper bound passes, we should minimize
    if maximize == minimize:
        raise ValueError("the decision boundary of the predicate is outside the bounds")

    if isinstance(lower, float):
        tolerance = 0.
        half = lambda x: x / 2.
    elif isinstance(lower, int):
        tolerance = 1  # the lower and upper bounds never meet due to int truncation
        half = lambda x: x // 2
    else:
        raise TypeError("bounds must be either float or int")  # pragma: no cover

    mid = lower
    while upper - lower > tolerance:
        new_mid = lower + half(upper - lower)  # avoid overflow

        # avoid an infinite loop from float roundoff
        if new_mid == mid:
            break

        mid = new_mid
        if predicate(mid) == minimize:
            upper = mid
        else:
            lower = mid

    # one bound is always false, the other true. Return the truthy bound
    value = upper if minimize else lower

    # optionally return sign
    if return_sign:
        return value, 1 if minimize else -1 # type: ignore
    
    return value


_EXPONENTIAL_SEARCH_BANDS: dict[Type, list[float]] = {
    # Searching bands of [(k - 1) * 2^16, k * 2^16].
    # Integers have linear space between bands.
    # Additionally include 1 because zero is prone to error.
    int: [0, 1, *(2 ** 16 * k for k in range(1, 9))],

    # Searching bands of [2^((k - 1)^2), 2^(k^2)].
    # Exponent has ten bits (2^1024 overflows) so k must be in [0, 32).
    # Unlikely to need numbers greater than 2**64, and to avoid overflow from shifted centers,
    #    only check k in [0, 8). 
    # Set your own bounds if this is not sufficient.
    float: [0.0, 0.5, *(2. ** k ** 2 for k in range(1024 // 32 // 4))]
}

def exponential_bounds_search(
    predicate: Callable[[float], bool], T: Optional[Union[Type[float], Type[int]]]
) -> Optional[tuple[float, float]]:
    """Determine bounds for a binary search via an exponential search,
    in large bands of [2^((k - 1)^2), 2^(k^2)] for k in [0, 8).
    Will attempt to recover once if `predicate` throws an exception, 
    by searching bands on the ok side of the exception boundary.
    

    :param predicate: a monotonic unary function from a number to a boolean
    :param T: type of argument to predicate, one of {float, int}
    :return: a tuple of float or int bounds that the decision boundary lies within
    :raises TypeError: if the type is not inferrable (pass T)
    :raises ValueError: if the predicate function is constant
    """

    # try to infer T
    T = T or _infer_type(predicate)
                
    # core search functionality
    def signed_band_search(center, at_center, sign):
        """Identify which band (of eight) the decision boundary lies in.

        :param center: Start here
        :param at_center: How the predicate evaluates at `center`. Search terminates when predicate changes
        :param sign: Search in this direction
        """

        if T not in _EXPONENTIAL_SEARCH_BANDS:
            raise TypeError(f"unknown type {T}. Must be one of int, float")  # pragma: no cover
        bands = [center + sign * c for c in _EXPONENTIAL_SEARCH_BANDS[T]]

        for i in range(1, len(bands)):
            # looking for a change in sign that indicates the decision boundary is within this band
            if at_center != predicate(bands[i]):
                # return the band
                return tuple(sorted(bands[i - 1:i + 1]))
        
        # No band found!
        return None

    center = {int: 0, float: 0.}[T]
    try:
        at_center = predicate(center)
        # search positive bands, then negative bands
        return signed_band_search(center, at_center, 1) or signed_band_search(center, at_center, -1)
    except Exception:
        pass

    # predicate has thrown an exception
    # 1. Treat exceptions as a secondary decision boundary, and find the edge value
    # 2. Return a bound by searching from the exception edge, in the direction away from the exception
    def _exception_predicate(v):
        try:
            predicate(v)
            return True
        except Exception:
            return False
    
    exception_bounds = exponential_bounds_search(_exception_predicate, T=T)
    if exception_bounds is None:
        try:
            predicate(center)
        except Exception as e:
            # enrich the error message if in Python 3.11+.
            if hasattr(e, "add_note"):
                e.add_note(f"Predicate in binary search always raises an exception. This exception is raised when the predicate is evaluated at {center}.")
            raise
    

    center, sign = binary_search(_exception_predicate, bounds=exception_bounds, T=T, return_sign=True)
    at_center = predicate(center)
    return signed_band_search(center, at_center, sign)


def _infer_type(predicate: Callable[[float], bool]) -> Union[Type[float], Type[int]]:
    def _is_type_error(e):
        is_match_error = isinstance(e, OpenDPException) and "No match for concrete type" in (e.message or "")
        is_type_error = isinstance(e, TypeError)
        return is_match_error or is_type_error

    try:
        predicate(0.)
        # predicate succeeded with a float, assume it wants floats
        return float
    except Exception as e:
        if not _is_type_error(e):
            # it didn't reject the type, but did fail with a different error
            return float
    
    # Try again with the more forgiving type, an int.
    try:
        predicate(0)
        # predicate is happy with an int, assume it wants ints
        return int
    except Exception as e:
        if not _is_type_error(e):
            # it didn't reject the type, but did fail with a different error
            return int

        # enrich the error message if in Python 3.11+.
        if hasattr(e, "add_note"):
            e.add_note("Unable to infer type `T`; pass the type `T` or bounds. This exception is raised when the predicate is evaluated at 0")
        raise


_TUPLE_FLAG = '__tuple__'
_EXPR_FLAG = '__expr__'
_LAZY_FLAG = '__lazyframe__'
_MARGIN_FLAG = '__margin__'
_FUNCTION_FLAG = '__function__'
_MODULE_FLAG = '__module__'
_KWARGS_FLAG = '__kwargs__'

def _bytes_to_b64_str(serialized_polars):
    from base64 import b64encode
    b64_bytes = b64encode(serialized_polars)
    return b64_bytes.decode()

def _b64_str_to_bytes(b64_str):
    from base64 import b64decode
    from io import BytesIO
    b64_bytes = b64_str.encode()
    serialized_polars = b64decode(b64_bytes)
    return BytesIO(serialized_polars)

class _Encoder(json.JSONEncoder):
    def default(self, obj):
        # Basic types:
        if isinstance(obj, (str, int, float, bool, type(None))):
            return obj
        if isinstance(obj, list):
            return [self.default(value) for value in obj]
        if isinstance(obj, tuple):
            return {_TUPLE_FLAG: [self.default(value) for value in obj]}
        if isinstance(obj, dict):
            return {
                # Dict keys must be hashable,
                # and with the JSON serialization we can't use tuple keys either,
                # so no need to recurse on the key.
                k: self.default(v)
                for k, v in obj.items()
            }
        
        from opendp.extras.polars import Margin
        if isinstance(obj, Margin):
            return {_MARGIN_FLAG: self.default(asdict(obj))}

        # OpenDP specific:
        if hasattr(obj, '__opendp_dict__'):
            return self.default({**obj.__opendp_dict__, '__version__': __version__})

        stateful_error_msg = (
            f"OpenDP JSON Encoder does not handle instances of {type(obj)}: "
            f"It may have state which is not set by the constructor. Error on: {obj}"
        )

        pl = import_optional_dependency('polars', raise_error=False)
        if pl is not None:
            from opendp.extras.polars import LazyFrameQuery
            if isinstance(obj, LazyFrameQuery):
                # Error out early, instead of falling through to pl.LazyFrame,
                # which *could* be serialized!
                raise Exception(stateful_error_msg)
            if isinstance(obj, pl.Expr):
                return {_EXPR_FLAG: _bytes_to_b64_str(obj.meta.serialize()), '__version__': __version__}
            if isinstance(obj, pl.LazyFrame):
                return {_LAZY_FLAG: _bytes_to_b64_str(obj.serialize()), '__version__': __version__}

        # Exceptions:
        from opendp.context import Context
        if isinstance(obj, (Context, Queryable,)):
            raise Exception(stateful_error_msg)

        raise Exception(f'OpenDP JSON Encoder does not handle {obj}')

def _check_version(dp_dict):
    from warnings import warn
    serialized_version = dp_dict['__version__']
    if serialized_version != __version__:
        warn(
            f'OpenDP version in serialized object ({serialized_version}) '
            f'!= this version ({__version__})')

def _deserialization_hook(dp_dict):
    if _FUNCTION_FLAG in dp_dict:
        _check_version(dp_dict)
        module = importlib.import_module(f"opendp.{dp_dict[_MODULE_FLAG]}")
        func = getattr(module, dp_dict[_FUNCTION_FLAG])
        return func(**dp_dict.get(_KWARGS_FLAG, {}))
    if _TUPLE_FLAG in dp_dict:
        return tuple(dp_dict[_TUPLE_FLAG])
    pl = import_optional_dependency('polars', raise_error=False)
    if pl is not None:
        if _EXPR_FLAG in dp_dict:
            _check_version(dp_dict)
            return pl.Expr.deserialize(_b64_str_to_bytes(dp_dict[_EXPR_FLAG]))
        if _LAZY_FLAG in dp_dict:
            _check_version(dp_dict)
            return pl.LazyFrame.deserialize(_b64_str_to_bytes(dp_dict[_LAZY_FLAG]))
        if _MARGIN_FLAG in dp_dict:
            from opendp.extras.polars import Margin

            by = [deserialize(v) for v in dp_dict[_MARGIN_FLAG].get('by', [])]
            return Margin(**{**dp_dict[_MARGIN_FLAG], "by": by})
    return dp_dict

def serialize(dp_obj):
    # The usual pattern would be
    # 
    #   json.dumps(dp_obj, cls=_Encoder)
    # 
    # but that only calls default() for objects which can't be handled otherwise.
    # In particular, it makes top-level tuples into lists,
    # even when special handling is specified in default().
    return json.dumps(_Encoder().default(dp_obj))

def deserialize(dp_json):
    return json.loads(dp_json, object_hook=_deserialization_hook)


_EXPECTED_POLARS_VERSION = '1.32.0' # Keep in sync with setup.cfg.


__version__ = get_opendp_version()



"""MST mechanism from `MMS21 <https://arxiv.org/abs/2108.04978>`_."""

from dataclasses import dataclass
import itertools
from typing import Any, Optional, cast, TYPE_CHECKING

from opendp._internal import _make_transformation, _new_pure_function
from opendp._lib import import_optional_dependency
from opendp.combinators import then_adaptive_composition
from opendp.domains import atom_domain, vector_domain
from opendp.extras.mbi._utilities import (
    TypedDictDistance,
    TypedDictDomain,
    get_associated_metric,
    make_noise_marginals,
    prior,
    make_stable_marginals,
    weight_marginals,
    Algorithm,
)
from opendp.measurements import make_noisy_max
from opendp.metrics import linf_distance
from opendp.mod import (
    ExtrinsicDistance,
    ExtrinsicDomain,
    FrameDistance,
    LazyFrameDomain,
    Measure,
    Measurement,
    Queryable,
    binary_search_chain,
)

if TYPE_CHECKING:  # pragma: no cover
    from opendp.extras.polars import Bound


@dataclass(kw_only=True, frozen=True)
class MST(Algorithm):
    """MST mechanism from `MMS21 <https://arxiv.org/abs/2108.04978>`_.
    
    MST greedily chooses pairs of columns that are most poorly represented 
    by the DP contingency table in a way that guarantees all columns become 
    connected by a minimum spanning tree.
    MST then releases all of the selected marginals.

    ..
        >>> import pytest  # `pip install opendp[mbi]` is necessary
        >>> _ = pytest.importorskip("mbi")
    
    .. code:: pycon

        >>> import opendp.prelude as dp
        >>> import polars as pl

        >>> dp.enable_features("contrib")

        >>> context = dp.Context.compositor(
        ...     data=pl.scan_csv(dp.examples.get_france_lfs_path(), ignore_errors=True),
        ...     privacy_unit=dp.unit_of(contributions=36),
        ...     privacy_loss=dp.loss_of(rho=0.1, delta=1e-7),
        ... )

        >>> table_mst = (
        ...     context.query(rho=0.1, delta=1e-7)
        ...     # transformations/truncation may be applied here
        ...     .select("SEX", "AGE", "HWUSUAL", "ILOSTAT")
        ...     .contingency_table(
        ...         keys={"SEX": [1, 2]}, 
        ...         cuts={"AGE": [20, 40, 60], "HWUSUAL": [1, 20, 40]},
        ...         algorithm=dp.mbi.MST()
        ...     )
        ...     .release()
        ... )

        >>> table_mst.synthesize() # doctest: +SKIP
        shape: (3_807_732, 4)
        ┌─────┬───────────┬───────────┬─────────┐
        │ SEX ┆ AGE       ┆ HWUSUAL   ┆ ILOSTAT │
        │ --- ┆ ---       ┆ ---       ┆ ---     │
        │ i64 ┆ f64       ┆ f64       ┆ i64     │
        ╞═════╪═══════════╪═══════════╪═════════╡
        │ 1   ┆ 55.446336 ┆ 20.776579 ┆ 1       │
        │ 1   ┆ 28.21838  ┆ 40.53348  ┆ 1       │
        │ 2   ┆ 43.291215 ┆ 34.406155 ┆ 1       │
        │ 1   ┆ 55.106615 ┆ 22.413161 ┆ 1       │
        │ 2   ┆ 42.585227 ┆ 40.11279  ┆ 3       │
        │ …   ┆ …         ┆ …         ┆ …       │
        │ 1   ┆ 58.197292 ┆ 40.139579 ┆ 1       │
        │ 1   ┆ 59.371221 ┆ 19.671153 ┆ 1       │
        │ 2   ┆ 19.862917 ┆ 40.339046 ┆ 9       │
        │ 1   ┆ 19.492355 ┆ 32.233661 ┆ 1       │
        │ 2   ┆ 60.863244 ┆ 40.737908 ┆ 3       │
        └─────┴───────────┴───────────┴─────────┘
    """

    measure_split: float = 0.9
    """Remaining proportion of budget to allocate to measuring marginals.
    
    The complement is spent on selecting marginals."""
    num_selections: Optional[int] = None
    """Number of second-order marginals to estimate.
    
    Defaults to one fewer than the number of columns in the data."""

    def __post_init__(self):
        super().__post_init__()

        if self.measure_split is not None and not (0 < self.measure_split <= 1):
            raise ValueError(f"measure_split ({self.measure_split}) must be in (0, 1]")

        if self.num_selections is not None and self.num_selections < 1:
            raise ValueError(f"num_selections ({self.num_selections}) must be positive")

    def make_marginals(
        self,
        input_domain: LazyFrameDomain,
        input_metric: FrameDistance,
        output_measure: Measure,
        d_in: list["Bound"],
        d_out: float,
        *,
        marginals: dict[tuple[str, ...], Any],
        model: Any,  # MarkovRandomField
    ) -> Measurement:
        """Implements MST (Minimum Spanning Tree) over ordinal data.

        :param input_domain: domain of input data
        :param input_metric: how to compute distance between datasets
        :param output_measure: how to measure privacy of release
        :param d_in: distance between adjacent input datasets
        :param d_out: upper bound on the privacy loss
        :param marginals: prior marginal releases
        :param model: warm-start fit of MarkovRandomField
        """
        import_optional_dependency("mbi")
        from mbi import MarkovRandomField  # type: ignore[import-untyped,import-not-found]

        if not isinstance(model, MarkovRandomField):
            raise ValueError("model must be a MarkovRandomField")

        d_measure = d_out * self.measure_split
        d_select = prior(prior(d_out - d_measure))

        lp_metric = get_associated_metric(output_measure)
        edges = list(itertools.combinations(input_domain.columns, 2))

        t_marginals = make_stable_marginals(input_domain, input_metric, lp_metric, edges)  # type: ignore[arg-type]
        d_marginals = t_marginals.map(d_in)

        def function(
            qbl: Queryable,
        ) -> tuple[dict[tuple[str, ...], Any], MarkovRandomField]:

            # SELECT a set of queries that best reduces the error
            m_select = _make_mst_select(
                *t_marginals.output_space,
                output_measure,
                d_in=d_marginals,
                d_out=d_select,
                edges=edges,
                model=model,
                num_selections=self.num_selections,
            )

            selected_cliques = qbl(m_select)

            # MEASURE selected marginals with noise
            m_measure = binary_search_chain(
                lambda s: make_noise_marginals(
                    *t_marginals.output_space,
                    output_measure,
                    selected_cliques,
                    scale=s,
                ),
                d_in=d_marginals,
                d_out=d_measure,
                T=float,
            )

            all_marginals = weight_marginals(marginals, *qbl(m_measure))

            # GENERATE (fit a MarkovRandomField)
            new_model = self.estimator(
                model.domain,
                list(all_marginals.values()),
                potentials=model.potentials.expand(model.cliques + selected_cliques),
            )
            return all_marginals, new_model

        return (
            t_marginals
            >> then_adaptive_composition(
                output_measure=output_measure,
                d_in=t_marginals.map(d_in),
                d_mids=[d_select, d_measure],
            )
            >> _new_pure_function(function)
        )


def _make_mst_select(
    input_domain: ExtrinsicDomain,
    input_metric: ExtrinsicDistance,
    output_measure: Measure,
    d_in,
    d_out,
    edges: list[tuple[str, str]],
    model,  # MarkovRandomField
    num_selections: Optional[int] = None,
) -> Measurement:
    """Make a measurement that selects a set of cliques that will minimize error."""
    from mbi import MarkovRandomField  # type: ignore[import-not-found]
    from scipy.cluster.hierarchy import DisjointSet  # type: ignore[import-untyped,import-not-found]

    model = cast(MarkovRandomField, model)

    max_selections = len(model.domain.attrs) - 1
    num_selections = min(max(0, num_selections or max_selections), max_selections)
    d_select = prior(d_out / num_selections)

    t_mst_scores = _make_mst_scores(input_domain, input_metric, edges, model)
    d_mst_scores = t_mst_scores.map(d_in)

    def function(qbl: Queryable) -> list[tuple[str, str]]:
        selected_edges = []
        components = DisjointSet(model.domain.attrs)

        for _ in range(num_selections):
            # filter down to only include edges that aren't connected
            unconnected = [
                i for i, e in enumerate(edges) if not components.connected(*e)
            ]

            t_unconnected = _make_transformation(  # type: ignore[misc]
                *t_mst_scores.output_space,
                *t_mst_scores.output_space,
                function=lambda scores: [scores[i] for i in unconnected],
                stability_map=lambda d_in: d_in,
            )

            m_rnm = binary_search_chain(
                lambda s: make_noisy_max(
                    *t_unconnected.output_space, output_measure=output_measure, scale=s
                ),
                d_in=d_mst_scores,
                d_out=d_select,
                T=float,
            )

            selected_edge = edges[unconnected[qbl(t_unconnected >> m_rnm)]]

            selected_edges.append(selected_edge)
            components.merge(*selected_edge)

        return selected_edges

    return (
        t_mst_scores
        >> then_adaptive_composition(
            output_measure,
            d_in=d_mst_scores,
            d_mids=[d_select] * num_selections,
        )
        >> _new_pure_function(function)
    )


def _make_mst_scores(
    input_domain: ExtrinsicDomain,
    input_metric: ExtrinsicDistance,
    queries: list[tuple[str, str]],
    model,  # MarkovRandomField
):
    """Make a transformation that assigns a score representing how poorly each query is estimated."""
    from opendp.extras.numpy import NPArrayDDomain
    import numpy as np  # type: ignore[import-not-found]
    from mbi import MarkovRandomField  # type: ignore[import-not-found]

    for value_domain in input_domain.cast(TypedDictDomain).values():
        value_domain.cast(NPArrayDDomain)

    input_metric.cast(TypedDictDistance)

    model = cast(MarkovRandomField, model)

    def score_query(query: tuple[str, str], exact) -> float:
        synth = model.project(query).values
        return float(np.linalg.norm(exact - synth, 1))

    return _make_transformation(
        input_domain=input_domain,
        input_metric=input_metric,
        output_domain=vector_domain(atom_domain(T="f64", nan=False)),
        output_metric=linf_distance(T="f64", monotonic=False),
        function=lambda exact_tabs: [
            score_query(query, exact_tabs[query]) for query in queries
        ],
        stability_map=lambda d_in: max(d_in.values()),
    )



# type: ignore
def top(
    iter: Iterator[T],
    k: usize,
    greater_than: Callable[[T, T], bool],
) -> list[T]:
    heap = []  # |\label{fn-heap}|
    
    if k == 0: # |\label{fn-zero}|
        return heap

    for value in iter:  # |\label{fn-iter}|
        if len(heap) == k: # |\label{fn-len}|
            if greater_than(heap[-1], value): # |\label{fn-greater}|
                continue
            heap.pop() # |\label{fn-pop}|

        # insert value into heap # |\label{fn-insert}|
        index = partition_point_mut(heap, lambda x: greater_than(x, value))
        heap.insert(index, value)
    return heap



"""
This module requires extra installs: ``pip install 'opendp[numpy]'``

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.numpy.canonical``.
"""

from __future__ import annotations

import math
from typing import Literal
from opendp._lib import import_optional_dependency
from opendp.mod import binary_search


class BinomialCND:
    """
    Utilities to conduct statistical inference on the output of the canonical noise mechanism on binomially-distributed data.

    Use :func:`opendp.measurements.make_canonical_noise` to instantiate the canonical noise mechanism.

    The mechanism outputs a sample from X + N,
    where X ~ Binomial(n=size, p=theta), N ~ CND(0, d_in, d_out).

    A counting query is a realization of X ~ Binomial(n=size, p=theta) when,
    for each row in data with ``size`` rows, a counter is incremented with probability theta.

    :param estimate: the differentially private outcome of the canonical noise mechanism on binomial data
    :param d_in: sensitivity of the input to the mechanism
    :param d_out: privacy parameters (ε, δ)
    :param size: (approximate) total number of records in the sensitive dataset
    """

    def __init__(
        self, estimate: float, d_in: float, d_out: tuple[float, float], size: int
    ) -> None:
        self.estimate = estimate
        self.d_in = d_in
        self.d_out = d_out
        self.size = size

    def confidence_interval(
        self, alpha: float, side: Literal["lower", "upper"] | None = None
    ) -> float | tuple[float, float]:
        """Compute a confidence interval for ``self.estimate``.

        * side=None: return the two bounds of a confidence interval
        * side="lower": return a one-sided confidence interval lower bound
        * side="upper": return a one-sided confidence interval upper bound

        :param alpha: statistical significance level
        :param side: configure interval type
        :return: confidence interval bound(s)
        """
        if side is None:
            return two_sided_confidence_interval(
                self.estimate, self.d_in, self.d_out, self.size, alpha
            )

        if side in {"lower", "upper"}:
            return one_sided_confidence_interval(
                self.estimate, self.d_in, self.d_out, self.size, alpha, side
            )

        raise ValueError(f"tail must be None, 'lower' or 'upper', not '{side}'")  # pragma: no cover

    def p_value(
        self, theta: float, tail: Literal["left", "right"] | None = None
    ) -> float:
        """
        Computes a p-value based on ``self.estimate``.

        * tail=None: null hypothesis states that true proportion is theta
        * tail="left": null hypothesis states that true proportion is no less than theta
        * tail="right": null hypothesis states that true proportion is no greater than theta

        :param theta: success rate
        :param tail: configure the null hypothesis
        :return: the probability of observing ``self.estimate``, given the null hypothesis
        """
        if tail is None:
            return two_sided_pvalue(
                self.estimate, self.d_in, self.d_out, self.size, theta
            )

        if tail in {"left", "right"}:
            return one_sided_pvalue(
                self.estimate, self.d_in, self.d_out, self.size, theta, tail
            )

        raise ValueError(f"tail must be None, 'left' or 'right', not '{tail}'")  # pragma: no cover

    def __repr__(self) -> str:
        return f"BinomialCND(estimate={self.estimate}, d_in={self.d_in}, d_out={self.d_out}, size={self.size})"


# If the canonical noise mechanism is generalized beyond (ε, δ)-DP, 
# then this needs to be updated, as it only considers the Tulap noise generated from this definition.
def _cnd_cdf(t, shift, d_in: float, d_out: tuple[float, float]):
    """Computes the value of the cumulative density function Pr[T <= t] where T ~ Tulap(shift, b, q),
    where b and q are calibrated to satisfy d_out-DP when sensitivity is d_in"""
    np = import_optional_dependency("numpy")
    epsilon, delta = d_out

    t = (np.atleast_1d(t) - shift) / d_in
    b = math.exp(-epsilon)
    q = (2 * delta * b) / (1 - b + 2 * delta * b)
    lcut = q / 2
    rcut = q / 2
    r = np.rint(t)
    g = -math.log(b)
    l = math.log(1 + b)  # noqa
    k = 1 - b
    
    with np.errstate(over="ignore"):
        negs = np.exp((r * g) - l + np.log(b + ((t - r + (1 / 2)) * k)))
        poss = 1 - np.exp((r * -g) - l + np.log(b + ((r - t + (1 / 2)) * k)))

    # check for infinities
    negs[np.isinf(negs)] = 0
    poss[np.isinf(poss)] = 0
    # truncate w.r.t. the indicator on t's positivity
    is_leq0 = np.less_equal(t, 0).astype(int)
    trunc = (is_leq0 * negs) + ((1 - is_leq0) * poss)

    # handle the cut adjustment and scaling
    q = lcut + rcut
    is_mid = np.logical_and(
        np.less_equal(lcut, trunc), np.less_equal(trunc, 1 - rcut)
    ).astype(int)
    is_rhs = np.less(1 - rcut, trunc).astype(int)
    return ((trunc - lcut) / (1 - q)) * is_mid + is_rhs


def two_sided_confidence_interval(
    Z: float, d_in: float, d_out: tuple[float, float], size: int, alpha: float
) -> tuple[float, float]:
    """Compute a two-sided confidence interval centered on ``Z``.

    :param Z: realization of cnd random variable
    :param d_in: sensitivity of the input to the mechanism
    :param d_out: privacy parameters (ε, δ)
    :param size: (approximate) total number of records in the sensitive dataset
    :param alpha: statistical significance level
    :return: confidence interval bounds
    """
    mle = max(min(Z / size, 1.0), 0.0)
    predicate = lambda B: two_sided_pvalue(B * size, d_in, d_out, size, mle) > alpha

    L = binary_search(predicate, bounds=(0.0, mle)) if mle > 0 else 0.0
    U = binary_search(predicate, bounds=(mle, 1.0)) if mle < 1 else 1.0

    return float(L), float(U)


def one_sided_confidence_interval(
    Z: float,
    d_in: float,
    d_out: tuple[float, float],
    size: int,
    alpha: float,
    side: Literal["lower", "upper"],
) -> float:
    """Compute a one-sided confidence interval for ``self.estimate``.

    * tail="lower": return a one-sided confidence interval lower bound
    * tail="upper": return a one-sided confidence interval upper bound

    :param Z: realization of cnd random variable
    :param d_in: sensitivity of the input to the mechanism
    :param d_out: privacy parameters (ε, δ)
    :param size: (approximate) total number of records in the sensitive dataset
    :param alpha: statistical significance level
    :param side: configure interval type
    :return: lower or upper confidence interval bound
    """
    mle = max(min(Z / size, 1.0), 0.0)
    tail = {"lower": "left", "upper": "right"}.get(side)
    if tail is None:
        raise ValueError(f"tail must be 'lower' or 'upper', not {tail}")  # pragma: no cover
    pred = lambda B: one_sided_pvalue(B * size, d_in, d_out, size, mle, tail) > alpha  # type: ignore[arg-type]
    return float(binary_search(pred, bounds=(0.0, 1.0)))


def two_sided_pvalue(
    Z: float, d_in: float, d_out: tuple[float, float], size: int, theta: float
) -> float:
    """Computes a p-value for a hypothesis test on the probability of success.

    :param Z: realization of cnd random variable
    :param d_in: sensitivity of the input to the mechanism
    :param d_out: privacy parameters (ε, δ)
    :param size: (approximate) number of trials
    :param theta: true probability of binomial distribution
    :return: Probability of observing Z when the true success rate is theta
    """
    T = abs(size * theta - Z)
    left_tail = one_sided_pvalue(
        Z=size * theta + T,
        d_in=d_in,
        d_out=d_out,
        size=size,
        theta=theta,
        tail="right",
    )
    right_tail = one_sided_pvalue(
        Z=size * theta - T,
        d_in=d_in,
        d_out=d_out,
        size=size,
        theta=theta,
        tail="right",
    )

    return float(left_tail + (1 - right_tail))


def one_sided_pvalue(
    Z: float,
    d_in: float,
    d_out: tuple[float, float],
    size: int,
    theta: float,
    tail: Literal["left", "right"],
) -> float:
    """Computes a p-value for a hypothesis test on the probability of success.

    * tail="left": null hypothesis states that true proportion is no less than theta
    * tail="right": null hypothesis states that true proportion is no greater than theta

    :param Z: realization of cnd random variable
    :param d_in: sensitivity of the input to the mechanism
    :param d_out: privacy parameters (ε, δ)
    :param size: (approximate) number of trials
    :param theta: true probability of binomial distribution
    :param tail: either "left" or "right"
    :return: Probability of observing Z when the true success rate is theta
    """
    stats = import_optional_dependency("scipy.stats")
    np = import_optional_dependency("numpy")

    values = np.arange(size)
    B = stats.binom.pmf(k=values, n=size, p=theta)
    if tail == "right":
        F = _cnd_cdf(values, Z, d_in=d_in, d_out=d_out)
    elif tail == "left":
        F = 1 - _cnd_cdf(values, Z, d_in=d_in, d_out=d_out)
    else:
        raise ValueError(f"tail must be 'left' or 'right', not '{tail}'")  # pragma: no cover
    return float(np.dot(F.T, B))


def one_sided_uniformly_most_powerful_tests(
    d_in: float, 
    d_out: tuple[float, float],
    size: int,
    theta: float,
    alpha: float,
    tail: Literal["left", "right"],
) -> list[float]:
    """Compute one-sided UMP tests for each choice of i in [size].

    * When tail="left", then the null hypothesis is: the estimate is at most ``i``
    * When tail="right", then the null hypothesis is: the estimate is at least ``i``

    :param d_in: sensitivity of the input to the mechanism
    :param d_out: privacy parameters (ε, δ)
    :param size: (approximate) number of trials
    :param theta: true probability of binomial distribution
    :param alpha: statistical significance level
    :param tail: either "left" or "right"
    :return: The probability of accepting the null hypothesis for each choice of i in [size].
    """
    stats = import_optional_dependency("scipy.stats")
    np = import_optional_dependency("numpy")

    values = np.arange(0, size + 1)
    B = stats.binom.pmf(k=values, n=size, p=theta)

    predicate = lambda s: np.dot(B, _cnd_cdf(values, s, d_in, d_out)) > alpha

    root = binary_search(predicate)
    phi = _cnd_cdf(values, root, d_in, d_out)

    if tail == "left":
        return phi

    if tail == "right":
        return 1 - phi

    raise ValueError(f"tail must be either 'left' or 'right', not '{tail}'")  # pragma: no cover



# type: ignore
def make_laplace_threshold(
    input_domain: DI,
    input_metric: MI,
    scale: f64,
    threshold: DI_Atom,
    k: Option[i32],
) -> Measurement[DI, DI_Carrier, MI, MO]:
    return DiscreteLaplace(scale, k).make_noise_threshold(
        (input_domain, input_metric), threshold
    )



# type: ignore
def score_candidates(
    x: Iterator[TIA], 
    candidates: list[TIA], 
    alpha_num: u64,
    alpha_den: u64,
    size_limit: u64
) -> Iterator[usize]:
    # count of the number of records between...
    #  (-inf, c1), [c1, c2), [c2, c3), ..., [ck, inf)
    hist_ro = [0] * candidates.len() + 1 # histogram of right-open intervals
    #  (-inf, c1], (c1, c2], (c2, c3], ..., (ck, inf)
    hist_lo = [0] * candidates.len() + 1 # histogram of left-open intervals

    for x_i in x:
        idx_lt = candidates.partition_point(lambda c: c < x_i)
        hist_lo[idx_lt] += 1  # `\label{hist-lo-increment}`

        idx_eq = idx_lt + candidates[idx_lt:].partition_point(lambda c: c == x_i)
        hist_ro[idx_eq] += 1  # `\label{hist-ro-increment}`

    n: u64 = hist_lo.iter().sum()  # `\label{n-sum}`

    # don't care about the number of elements greater than all candidates
    hist_ro.pop()  # `\label{pop-ro}`
    hist_lo.pop()  # `\label{pop-lo}`

    lt, le = 0, 0
    for ro, lo in zip(hist_ro, hist_lo):  # `\label{zip-hist}`
        # cumsum the right-open histogram to get the total number of records less than the candidate
        lt += ro  # `\label{lt-cumsum}`
        # cumsum the right-open histogram to get the total number of records lt or equal to the candidate
        le += lo  # `\label{le-cumsum}`

        gt = n - le # `\label{gt}`

        # the number of records equal to the candidate is the difference between the two cumsums
        lt_lim, gt_lim = lt.min(size_limit), gt.min(size_limit)  # `\label{lt-gt-lim}`

        # a_den * |    (1 - a)         * #(x < c)    -            a * #(x > c)|
        yield ((alpha_den - alpha_num) * lt_lim).abs_diff(alpha_num * gt_lim)  # `\label{score}`
